import grpc
import json
import logging
import os
import sys
import pickle
import datetime
from urllib import parse
import d3m_automl_rpc.core_pb2 as pb_core
import d3m_automl_rpc.core_pb2_grpc as pb_core_grpc
import d3m_automl_rpc.value_pb2 as pb_value
from d3m_automl_rpc.utils import encode_problem_description, encode_performance_metric, encode_value
from alphad3m.grpc_api.grpc_logger import LoggingStub


logger = logging.getLogger(__name__)


def do_hello(core):
    core.Hello(pb_core.HelloRequest())


def do_listprimitives(core):
    core.ListPrimitives(pb_core.ListPrimitivesRequest())


def do_search(core, problem, dataset_path, time_bound=30.0, pipelines_limit=0, pipeline_template=None):
    version = pb_core.DESCRIPTOR.GetOptions().Extensions[pb_core.protocol_version]
    automl_hyperparameters = {'exclude_primitives': None,
                              'include_primitives': None}
    automl_hyperparams_encoded = {k: encode_value({'type': 'object', 'value': v}, ['RAW'], '/tmp') for k, v in
                                  automl_hyperparameters.items()}
    search = core.SearchSolutions(pb_core.SearchSolutionsRequest(
        user_agent='ta3_stub',
        version=version,
        time_bound_search=time_bound,
        rank_solutions_limit=pipelines_limit,
        allowed_value_types=['CSV_URI'],
        problem=encode_problem_description(problem),
        automl_hyperparameters=automl_hyperparams_encoded,
        template=pipeline_template,
        inputs=[pb_value.Value(
            dataset_uri='file://%s' % dataset_path,
        )],
    ))

    start_time = datetime.datetime.now()
    results = core.GetSearchSolutionsResults(
        pb_core.GetSearchSolutionsResultsRequest(
            search_id=search.search_id,
        )
    )
    solutions = {}
    for result in results:
        if result.solution_id:
            end_time = datetime.datetime.now()
            solutions[result.solution_id] = (
                result.internal_score,
                result.scores,
                str(end_time - start_time)
            )

    return str(search.search_id), solutions


def do_score(core, problem, solutions, dataset_path):
    metrics = []

    for metric in problem['problem']['performance_metrics']:
        metrics.append(encode_performance_metric(metric))

    for solution in solutions:
        try:
            response = core.ScoreSolution(pb_core.ScoreSolutionRequest(
                solution_id=solution,
                inputs=[pb_value.Value(
                    dataset_uri='file://%s' % dataset_path,
                )],
                performance_metrics=metrics,
                users=[],
                configuration=pb_core.ScoringConfiguration(
                    method='K_FOLD',
                    folds=4,
                    train_test_ratio=0.75,
                    shuffle=True,
                    random_seed=0
                ),
            ))
            results = core.GetScoreSolutionResults(
                pb_core.GetScoreSolutionResultsRequest(
                    request_id=response.request_id,
                )
            )
            for _ in results:
                pass
        except Exception:
            logger.exception("Exception during scoring %r", solution)


def do_train(core, solutions, dataset_path):
    fitted = {}
    for solution in solutions:
        try:
            response = core.FitSolution(pb_core.FitSolutionRequest(
                solution_id=solution,
                inputs=[pb_value.Value(
                    dataset_uri='file://%s' % dataset_path,
                )],
                expose_outputs=['outputs.0'],
                expose_value_types=['CSV_URI'],
                users=[],
            ))
            results = core.GetFitSolutionResults(
                pb_core.GetFitSolutionResultsRequest(
                    request_id=response.request_id,
                )
            )
            for result in results:
                if result.progress.state == pb_core.COMPLETED:
                    fitted[solution] = result.fitted_solution_id
        except Exception:
            logger.exception("Exception training %r", solution)
    return fitted


def do_test(core, fitted, dataset_path):
    tested = {}
    for fitted_solution in fitted.values():
        try:
            response = core.ProduceSolution(pb_core.ProduceSolutionRequest(
                fitted_solution_id=fitted_solution,
                inputs=[pb_value.Value(
                    dataset_uri='file://%s' % dataset_path,
                )],
                expose_outputs=['outputs.0'],
                expose_value_types=['CSV_URI'],
                users=[],
            ))
            results = core.GetProduceSolutionResults(
                pb_core.GetProduceSolutionResultsRequest(
                    request_id=response.request_id,
                )
            )
            for result in results:
                if result.progress.state == pb_core.COMPLETED:
                    tested[fitted_solution] = result.exposed_outputs['outputs.0'].csv_uri
        except Exception:
            logger.exception("Exception testing %r", fitted_solution)

    return tested


def do_export(core, fitted):
    for i, fitted_solution in enumerate(fitted.values()):
        try:
            core.SolutionExport(pb_core.SolutionExportRequest(
                solution_id=fitted_solution,
                rank=(i + 1.0) / (len(fitted) + 1.0),
            ))
        except Exception:
            logger.exception("Exception exporting %r", fitted_solution)


def do_describe(core, solutions):
    for solution in solutions:
        try:
            core.DescribeSolution(pb_core.DescribeSolutionRequest(
                solution_id=solution,
            ))
        except Exception:
            logger.exception("Exception during describe %r", solution)


def do_save_solution(core, solution_id):
    response = core.SaveSolution(pb_core.SaveSolutionRequest(solution_id=solution_id))

    return response.solution_uri


def do_load_solution(core, solution_path):
    solution_uri = 'file://%s' % solution_path
    response = core.LoadSolution(pb_core.LoadSolutionRequest(solution_uri=solution_uri))

    return response.solution_id


def do_save_fitted_solution(core, fitted):
    saved = {}

    for fitted_solution_id in fitted.values():
        response = core.SaveFittedSolution(pb_core.SaveFittedSolutionRequest(fitted_solution_id=fitted_solution_id))
        parsed_uri = parse.urlparse(response.fitted_solution_uri, allow_fragments=False)
        fitted_solution_path = parsed_uri.path

        with open(fitted_solution_path, 'rb') as fin:
            fitted_object = pickle.load(fin)

        saved[fitted_solution_id] = fitted_object

    return saved


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(message)s")

    channel = grpc.insecure_channel('localhost:{0}'.format(os.environ.get('D3MPORT', 45042)))
    core = LoggingStub(pb_core_grpc.CoreStub(channel), logger)
    train_dataset_path = '/input/TRAIN/dataset_TRAIN/datasetDoc.json'
    test_dataset_path = '/input/TEST/dataset_TEST/datasetDoc.json'

    with open(sys.argv[1]) as problem:
        problem = json.load(problem)

    # Do a hello
    do_hello(core)

    # Do a list primitives
    do_listprimitives(core)
    # Do a search
    solutions = do_search(core, problem, train_dataset_path)

    # Describe the pipelines
    do_describe(core, solutions)

    # Score all found solutions
    do_score(core, problem, solutions, train_dataset_path)

    # Train all found solutions
    fitted = do_train(core, solutions, train_dataset_path)

    # Test all fitted solutions
    do_test(core, fitted, test_dataset_path)

    # Export all fitted solutions
    do_export(core, fitted)


if __name__ == '__main__':
    main()



"""GRPC server code, exposing AlphaD3M over the TA3-TA2 protocol.

Those adapters wrap the D3mTa2 object and handle all the GRPC and protobuf
logic, converting to/from protobuf messages. No GRPC or protobuf objects should
leave this module.
"""


import grpc
import json
import calendar
import datetime
import logging
from uuid import UUID
from os.path import join
from alphad3m import __version__
import d3m_automl_rpc.core_pb2 as pb_core
import d3m_automl_rpc.core_pb2_grpc as pb_core_grpc
import d3m_automl_rpc.problem_pb2 as pb_problem
import d3m_automl_rpc.value_pb2 as pb_value
import d3m_automl_rpc.pipeline_pb2 as pb_pipeline
import d3m_automl_rpc.primitive_pb2 as pb_primitive
from d3m_automl_rpc.utils import decode_pipeline_description, decode_problem_description, decode_performance_metric, \
    encode_pipeline_description, decode_value
from d3m.metadata import pipeline as pipeline_module
from d3m.metadata.problem import Problem
from google.protobuf.timestamp_pb2 import Timestamp
from alphad3m.grpc_api.grpc_logger import log_service
from alphad3m.primitive_loader import load_primitives_list
from alphad3m.utils import PersistentQueue

logger = logging.getLogger(__name__)


def to_timestamp(dt):
    """Converts a UTC datetime object into a gRPC Timestamp.

    :param dt: Time to convert, or None for now.
    :type dt: datetime.datetime | None
    """
    if dt is None:
        dt = datetime.datetime.utcnow()
    return Timestamp(seconds=calendar.timegm(dt.timetuple()),
                     nanos=dt.microsecond * 1000)


def error(context, code, format, *args):
    message = format % args
    context.set_code(code)
    context.set_details(message)
    if code == grpc.StatusCode.NOT_FOUND:
        return KeyError(message)
    else:
        return ValueError(message)


@log_service(logger)
class CoreService(pb_core_grpc.CoreServicer):

    def __init__(self, ta2):
        self._ta2 = ta2
        self._ta2.add_observer(self._ta2_event)
        self._requests = {}

    def _ta2_event(self, event, **kwargs):
        if 'job_id' in kwargs and kwargs['job_id'] in self._requests:
            job_id = kwargs['job_id']
            self._requests[job_id].put((event, kwargs))
            if event in ('scoring_success', 'scoring_error',
                         'training_success', 'training_error',
                         'test_success', 'test_error'):
                self._requests[job_id].close()

    def Hello(self, request, context):
        version = pb_core.DESCRIPTOR.GetOptions().Extensions[pb_core.protocol_version]
        user_agent = "alphad3m %s" % __version__

        return pb_core.HelloResponse(
            user_agent=user_agent,
            version=version,
            allowed_value_types=['RAW', 'DATASET_URI', 'CSV_URI'],
            supported_extensions=[],
            supported_task_keywords=[],  # TODO: Add supported_task_keywords using core package enums
            supported_performance_metrics=[],  # TODO: Add supported_performance_metrics using core package enums
            supported_evaluation_methods=['K_FOLD', 'HOLDOUT', 'RANKING'],
            supported_search_features=[]
        )

    def SearchSolutions(self, request, context):
        """Create a `Session` and start generating & scoring pipelines.
        """
        if len(request.inputs) > 1:
            raise error(context, grpc.StatusCode.UNIMPLEMENTED,
                        "Search with more than 1 input is not supported")
        expected_version = pb_core.DESCRIPTOR.GetOptions().Extensions[
            pb_core.protocol_version]

        if request.version != expected_version:
            logger.error("TA3 is using a different protocol version: %r "
                         "(us: %r)", request.version, expected_version)

        template = request.template

        if template is not None and len(template.steps) > 0:  # isinstance(template, pb_pipeline.PipelineDescription)
            pipeline = decode_pipeline_description(template, pipeline_module.Resolver())
            if pipeline.has_placeholder():
                template = pipeline.to_json_structure()
            else:  # Pipeline template fully defined
                problem = None
                if request.problem:
                    problem = decode_problem_description(request.problem)
                search_id = self._ta2.new_session(problem)
                dataset = request.inputs[0].dataset_uri
                if not dataset.startswith('file://'):
                    dataset = 'file://' + dataset

                self._ta2.build_fixed_pipeline(search_id, pipeline.to_json_structure(), dataset)
                return pb_core.SearchSolutionsResponse(search_id=str(search_id),)
        else:
            template = None

        dataset = request.inputs[0].dataset_uri
        if not dataset.endswith('datasetDoc.json'):
            raise error(context, grpc.StatusCode.INVALID_ARGUMENT,
                        "Dataset is not in D3M format: %s", dataset)
        if not dataset.startswith('file://'):
            dataset = 'file://' + dataset

        problem = decode_problem_description(request.problem)
        timeout_search = request.time_bound_search
        timeout_run = request.time_bound_run
        report_rank = True if request.rank_solutions_limit > 0 else False

        if timeout_search <= 0.0: timeout_search = None
        if timeout_run <= 0.0: timeout_run = None

        search_id = self._ta2.new_session(problem)
        session = self._ta2.sessions[search_id]
        task_keywords = session.problem['problem']['task_keywords']
        metrics = session.metrics

        automl_hyperparameters = {}
        if request.automl_hyperparameters:
            for hp_name, hp_value in request.automl_hyperparameters.items():
                automl_hyperparameters[hp_name] = decode_value(hp_value)['value']

        self._ta2.build_pipelines(search_id, dataset, task_keywords, metrics, timeout_search, timeout_run,
                                  automl_hyperparameters, template, report_rank=report_rank)

        return pb_core.SearchSolutionsResponse(
            search_id=str(search_id),
        )

    def GetSearchSolutionsResults(self, request, context):
        """Get the created pipelines and scores.
        """
        session_id = UUID(hex=request.search_id)
        if session_id not in self._ta2.sessions:
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown search ID %r", session_id)

        session = self._ta2.sessions[session_id]

        def msg_solution(pipeline_id):
            scores = self._ta2.get_pipeline_scores(pipeline_id)
            progress = session.progress

            if scores:
                if session.metrics and session.metrics[0]['metric'].name in scores:
                    metric = session.metrics[0]['metric']
                    try:
                        internal_score = metric.normalize(scores[metric.name])
                    except:
                        internal_score = scores[metric.name]
                        logger.warning('Problems normalizing metric, using the raw value: %.2f' % scores[metric.name])
                else:
                    internal_score = float('nan')
                scores = [
                    pb_core.Score(
                        metric=pb_problem.ProblemPerformanceMetric(
                            metric=m,
                            k=0,
                            pos_label=''),
                        value=pb_value.Value(
                            raw=pb_value.ValueRaw(double=s)
                        ),
                    )
                    for m, s in scores.items()
                ]
                scores = [pb_core.SolutionSearchScore(scores=scores)]
                return pb_core.GetSearchSolutionsResultsResponse(
                    done_ticks=progress.current,
                    all_ticks=progress.total,
                    progress=pb_core.Progress(
                        state=pb_core.RUNNING,
                        status="Solution scored",
                        start=to_timestamp(session.start),
                    ),
                    solution_id=str(pipeline_id),
                    internal_score=internal_score,
                    scores=scores,
                )

        def msg_progress(status, state=pb_core.RUNNING):
            progress = session.progress

            return pb_core.GetSearchSolutionsResultsResponse(
                done_ticks=progress.current,
                all_ticks=progress.total,
                progress=pb_core.Progress(
                    state=state,
                    status=status,
                    start=to_timestamp(session.start),
                ),
                internal_score=float('nan'),
            )

        def msg_fixed_solution(pipeline_id, state=pb_core.RUNNING):
            progress = session.progress

            return pb_core.GetSearchSolutionsResultsResponse(
                done_ticks=progress.current,
                all_ticks=progress.total,
                progress=pb_core.Progress(
                    state=state,
                    status="Solution Created",
                    start=to_timestamp(session.start),
                ),
                solution_id=str(pipeline_id),
                internal_score=float('nan'),
            )

        with session.with_observer_queue() as queue:
            # Send the solutions that already exist
            for pipeline_id in session.pipelines:
                msg = msg_solution(pipeline_id)
                if msg is not None:
                    yield msg

            # Send updates by listening to notifications on session
            while session.working or not queue.empty():
                if not context.is_active():
                    logger.info(
                        "Client closed GetSearchSolutionsResults stream")
                    break
                event, kwargs = queue.get()
                if event == 'finish_session' or event == 'done_searching':
                    break
                elif event == 'new_pipeline':
                    yield msg_progress("Trying new solution")
                elif event == 'new_fixed_pipeline':
                    pipeline_id = kwargs['pipeline_id']
                    yield msg_fixed_solution(pipeline_id)
                elif event == 'scoring_success':
                    pipeline_id = kwargs['pipeline_id']
                    msg = msg_solution(pipeline_id)
                    if msg is not None:
                        yield msg
                    else:
                        yield msg_progress("No appropriate score")
                elif event == 'scoring_error':
                    yield msg_progress("Solution doesn't work")

            yield msg_progress("End of search", pb_core.COMPLETED)

    def ScoreSolution(self, request, context):
        """Request scores for a pipeline.
        """
        pipeline_id = UUID(hex=request.solution_id)

        dataset = request.inputs[0].dataset_uri
        if not dataset.endswith('datasetDoc.json'):
            raise error(context, grpc.StatusCode.INVALID_ARGUMENT,
                        "Dataset is not in D3M format: %s", dataset)

        if dataset.startswith('/'):
            logger.warning("Dataset is a path, turning it into a file:// URL")
            dataset = 'file://' + dataset

        metrics = []
        for metric in request.performance_metrics:
                metrics.append(decode_performance_metric(metric))

        logger.info("Got ScoreSolution request, dataset=%s, "
                    "metrics=%s",
                    dataset, metrics)

        problem = None
        timeout_run = None
        for session_id in self._ta2.sessions.keys():
            if pipeline_id in self._ta2.sessions[session_id].pipelines:
                problem = self._ta2.sessions[session_id].problem
                timeout_run = self._ta2.sessions[session_id].timeout_run
                break

        scoring_config = {
                        'method': request.configuration.method,
                        'train_score_ratio': str(request.configuration.train_test_ratio),
                        'random_seed': request.configuration.random_seed,
                        'shuffle': str(request.configuration.shuffle).lower(),
                        'stratified': str(request.configuration.stratified).lower()
                        }
        if scoring_config['method'] == 'K_FOLD':
            scoring_config['number_of_folds'] = str(request.configuration.folds)

        job_id = self._ta2.score_pipeline(pipeline_id, metrics, dataset, problem, scoring_config, timeout_run)
        self._requests[job_id] = PersistentQueue()

        return pb_core.ScoreSolutionResponse(
            request_id='%x' % job_id,
        )

    def GetScoreSolutionResults(self, request, context):
        """Wait for a scoring job to be done.
        """
        try:
            job_id = int(request.request_id, 16)
            queue = self._requests[job_id]
        except (ValueError, KeyError):
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown ID %r", request.request_id)

        for event, kwargs in queue.read():
            if not context.is_active():
                logger.info("Client closed GetScoreSolutionResults stream")
                break

            if event == 'scoring_start':
                yield pb_core.GetScoreSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.RUNNING,
                        status="Scoring in progress",
                    ),
                )
            elif event == 'scoring_success':
                pipeline_id = kwargs['pipeline_id']
                scores = self._ta2.get_pipeline_scores(pipeline_id)
                scores = [
                    pb_core.Score(
                        metric=pb_problem.ProblemPerformanceMetric(
                            metric=m,
                            k=0,
                            pos_label=''),
                        value=pb_value.Value(
                            raw=pb_value.ValueRaw(double=s)
                        ),
                    )
                    for m, s in scores.items()
                ]
                yield pb_core.GetScoreSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.COMPLETED,
                        status="Scoring completed",
                    ),
                    scores=scores,
                )
                break
            elif event == 'scoring_error':
                status = kwargs['error_msg']
                yield pb_core.GetScoreSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.ERRORED,
                        status=status,
                    ),
                )
                break

    def FitSolution(self, request, context):
        """Train a pipeline on a dataset.

        This will make it available for testing and exporting.
        """
        pipeline_id = UUID(hex=request.solution_id)
        dataset = request.inputs[0].dataset_uri
        steps_to_expose = list(request.expose_outputs)

        if dataset.startswith('/'):
            logger.warning("Dataset is a path, turning it into a file:// URL")
            dataset = 'file://' + dataset

        problem = None
        for session_id in self._ta2.sessions.keys():
            if pipeline_id in self._ta2.sessions[session_id].pipelines:
                problem = self._ta2.sessions[session_id].problem
                break

        job_id = self._ta2.train_pipeline(pipeline_id, dataset, problem, steps_to_expose)
        self._requests[job_id] = PersistentQueue()

        return pb_core.FitSolutionResponse(
            request_id='%x' % job_id,
        )

    def GetFitSolutionResults(self, request, context):
        """Wait for a training job to be done.
        """
        try:
            job_id = int(request.request_id, 16)
            queue = self._requests[job_id]
        except (ValueError, KeyError):
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown ID %r", request.request_id)

        for event, kwargs in queue.read():
            if not context.is_active():
                logger.info("Client closed GetFitSolutionsResults stream")
                break

            if event == 'training_start':
                yield pb_core.GetFitSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.RUNNING,
                        status="Training in progress",
                    ),
                )
            elif event == 'training_success':
                pipeline_id = kwargs['pipeline_id']
                storage_dir = kwargs['storage_dir']
                steps_to_expose = kwargs['steps_to_expose']
                yield pb_core.GetFitSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.COMPLETED,
                        status="Training completed",
                    ),
                    exposed_outputs={step_id: pb_value.Value(csv_uri='file://%s/fit_%s_%s.csv' %
                                                                     (storage_dir, pipeline_id, step_id))
                                     for step_id in steps_to_expose},
                    fitted_solution_id=str(pipeline_id),
                )
                break
            elif event == 'training_error':
                status = kwargs['error_msg']
                yield pb_core.GetFitSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.ERRORED,
                        status=status,
                    ),
                )
                break
            elif event == 'done_searching':
                break

    def ProduceSolution(self, request, context):
        """Run testing from a trained pipeline.
        """
        pipeline_id = UUID(hex=request.fitted_solution_id)
        dataset = request.inputs[0].dataset_uri
        steps_to_expose = list(request.expose_outputs)

        if dataset.startswith('/'):
            logger.warning("Dataset is a path, turning it into a file:// URL")
            dataset = 'file://' + dataset

        job_id = self._ta2.test_pipeline(pipeline_id, dataset, steps_to_expose)
        self._requests[job_id] = PersistentQueue()

        return pb_core.ProduceSolutionResponse(
            request_id='%x' % job_id,
        )

    def GetProduceSolutionResults(self, request, context):
        """Wait for the requested test run to be done.
        """
        try:
            job_id = int(request.request_id, 16)
            queue = self._requests[job_id]
        except (ValueError, KeyError):
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown ID %r", request.request_id)

        for event, kwargs in queue.read():
            if not context.is_active():
                logger.info("Client closed GetProduceSolutionResults "
                            "stream")
                break
            if event == 'testing_success':
                pipeline_id = kwargs['pipeline_id']
                storage_dir = kwargs['storage_dir']
                steps_to_expose = kwargs['steps_to_expose']
                yield pb_core.GetProduceSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.COMPLETED,
                        status="Execution completed",
                    ),
                    exposed_outputs={step_id: pb_value.Value(csv_uri='file://%s/produce_%s_%s.csv' %
                                                                     (storage_dir, pipeline_id, step_id))
                                     for step_id in steps_to_expose},
                )
                break
            elif event == 'testing_error':
                status = kwargs['error_msg']
                yield pb_core.GetProduceSolutionResultsResponse(
                    progress=pb_core.Progress(
                        state=pb_core.ERRORED,
                        status=status,
                    ),
                )
                break

    def SaveSolution(self, request, context):
        """Save a pipeline.
        """
        pipeline_id = UUID(hex=request.solution_id)
        session_id = None
        problem = None

        for session_id in self._ta2.sessions.keys():
            if pipeline_id in self._ta2.sessions[session_id].pipelines:
                problem = self._ta2.sessions[session_id].problem
                break

        solution_uri = self._ta2.save_pipeline(session_id, pipeline_id, problem)

        return pb_core.SaveSolutionResponse(
            solution_uri='%s' % solution_uri,
        )

    def LoadSolution(self, request, context):
        """Load a pipeline into a search session.
        """
        solution_path = request.solution_uri.replace('file://', '')

        with open(join(solution_path, 'pipeline.json')) as fin:
            pipeline = json.load(fin)

        with open(join(solution_path, 'problem.json')) as fin:
            problem_json = json.load(fin)
            problem = Problem.from_json_structure(problem_json)

        search_id = self._ta2.new_session(problem)
        solution_id = self._ta2.load_pipeline(search_id, pipeline)

        return pb_core.LoadSolutionResponse(
            solution_id='%s' % solution_id,
        )

    def SaveFittedSolution(self, request, context):
        """Save a trained pipeline.
        """
        pipeline_id = UUID(hex=request.fitted_solution_id)
        fitted_solution_uri = self._ta2.save_fitted_pipeline(str(pipeline_id))

        return pb_core.SaveFittedSolutionResponse(
            fitted_solution_uri='%s' % fitted_solution_uri,
        )

    def DescribeSolution(self, request, context):
        """Describe a pipeline as a grpc message.
        """
        pipeline_id = UUID(hex=request.solution_id)
        pipeline = self._ta2.get_pipeline(pipeline_id)

        if not pipeline:
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown solution ID %r", request.solution_id)

        grpc_pipeline = encode_pipeline_description(pipeline, ['RAW', 'DATASET_URI', 'CSV_URI'], self._ta2.output_folder)

        return pb_core.DescribeSolutionResponse(
            pipeline=grpc_pipeline,
            steps=None
        )

    def SolutionExport(self, request, context):
        """Export a pipeline.
        """
        pipeline_id = UUID(hex=request.solution_id)
        session_id = None
        for session_key in self._ta2.sessions:
            session = self._ta2.sessions[session_key]
            with session.lock:
                if pipeline_id in session.pipelines:
                    session_id = session.id
                    break
        if session_id is None:
            raise error(context, grpc.StatusCode.NOT_FOUND,
                        "Unknown solution ID %r", request.solution_id)
        session = self._ta2.sessions[session_id]
        rank = request.rank
        if rank < 0.0:
            rank = None

        session.write_exported_pipeline(pipeline_id, rank)
        return pb_core.SolutionExportResponse()

    def ListPrimitives(self, request, context):
        primitives = []
        installed_primitives = load_primitives_list()

        for primitive in installed_primitives:
            primitives.append(pb_primitive.Primitive(id=primitive['id'], version=primitive['version'],
                                                     python_path=primitive['python_path'], name=primitive['name'],
                                                     digest=primitive['digest']))

        return pb_core.ListPrimitivesResponse(primitives=primitives)

    def StopSearchSolutions(self, request, context):
        """Stop the search without deleting the `Session`.
        """
        session_id = UUID(hex=request.search_id)
        if session_id in self._ta2.sessions:
            self._ta2.stop_session(session_id)
            logger.info("Search stopped: %s", session_id)
        return pb_core.StopSearchSolutionsResponse()

    def EndSearchSolutions(self, request, context):
        """Stop the search and delete the `Session`.
        """
        session_id = UUID(hex=request.search_id)
        if session_id in self._ta2.sessions:
            self._ta2.finish_session(session_id)
            logger.info("Search terminated: %s", session_id)
        return pb_core.EndSearchSolutionsResponse()



"""GRPC logger, utility to log all messages.
"""

import collections
import functools
import string


def log_message_out(logger, msg, name):
    logger.info("< %s", name)
    for line in str(msg).splitlines():
        logger.info("< | %s", line)
    logger.info("  ------------------")


def log_message_in(logger, msg, name):
    logger.info("> %s", name)
    for line in str(msg).splitlines():
        logger.info("> | %s", line)
    logger.info("  ------------------")


def log_stream_out(logger, gen, name):
    for msg in gen:
        log_message_out(logger, msg, name)
        yield msg


def log_stream_in(logger, gen, name):
    for msg in gen:
        log_message_in(logger, msg, name)
        yield msg


def _wrap(logger, func):
    name = func.__name__

    @functools.wraps(func)
    def wrapped(self, request, context):
        log_message_in(logger, request, name)
        ret = func(self, request, context)
        if isinstance(ret, collections.Iterable):
            return log_stream_out(logger, ret, name)
        else:
            log_message_out(logger, ret, name)
            return ret

    return wrapped


def log_service(logger):
    def wrapper(klass):
        base, = klass.__bases__
        for name in dir(base):
            if name[0] not in string.ascii_uppercase:
                continue
            if name not in klass.__dict__:
                continue
            setattr(klass, name, _wrap(logger, klass.__dict__[name]))
        return klass
    return wrapper


class LoggingStub(object):
    def __init__(self, stub, logger):
        self._stub = stub
        self._logger = logger

    def __getattr__(self, item):
        return self._wrap(getattr(self._stub, item), item)

    def _wrap(self, method, name):
        def wrapper(out_msg):
            log_message_out(self._logger, out_msg, name)

            in_msg = method(out_msg)

            if isinstance(in_msg, collections.Iterable):
                return log_stream_in(self._logger, in_msg, name)
            else:
                log_message_in(self._logger, in_msg, name)
                return in_msg

        return wrapper



"""Various utilities that are not specific to D3M.
"""

import os
import contextlib
import logging
import json
from queue import Empty, Queue
import threading
import subprocess
from d3m.metadata.problem import TaskKeyword
from sklearn.model_selection import train_test_split
from alphad3m.primitive_loader import load_primitives_list

SAMPLE_SIZE = 2000
RANDOM_SEED = 0

logger = logging.getLogger(__name__)


class Observable(object):
    """Allow adding callbacks on an object, to be called on notifications.
    """
    def __init__(self):
        self.__observers = {}
        self.__next_key = 0
        self.lock = threading.RLock()

    def add_observer(self, observer):
        with self.lock:
            key = self.__next_key
            self.__next_key += 1
            self.__observers[key] = observer
            return key

    def remove_observer(self, key):
        with self.lock:
            del self.__observers[key]

    @contextlib.contextmanager
    def with_observer(self, observer):
        key = self.add_observer(observer)
        try:
            yield
        finally:
            self.remove_observer(key)

    @contextlib.contextmanager
    def with_observer_queue(self):
        queue = Queue()
        with self.with_observer(lambda e, **kw: queue.put((e, kw))):
            yield queue

    def notify(self, event, **kwargs):
        with self.lock:
            for observer in self.__observers.values():
                try:
                    observer(event, **kwargs)
                except Exception:
                    logging.exception("Error in observer")


class _PQ_Reader(object):
    def __init__(self, pq):
        self._pq = pq
        self._pos = 0
        self.finished = False

    def get(self, timeout=None):
        if self.finished:
            return None
        with self._pq.lock:
            # There are unread items
            if (len(self._pq.list) > self._pos or
                    # Or get woken up
                    self._pq.change.wait(timeout)):
                self._pos += 1
                item = self._pq.list[self._pos - 1]
                if item is None:
                    self.finished = True
                return item
            # Timeout
            else:
                raise Empty


class PersistentQueue(object):
    """A Queue object that will always yield items inserted from the start.
    """
    def __init__(self):
        self.list = []
        self.lock = threading.RLock()
        self.change = threading.Condition(self.lock)

    def put(self, item):
        """Put an item in the queue, waking up readers.
        """
        if item is None:
            raise TypeError("Can't put None in PersistentQueue")
        with self.lock:
            self.list.append(item)
            self.change.notify_all()

    def close(self):
        """End the queue, readers will terminate.
        """
        with self.lock:
            self.list.append(None)
            self.change.notify_all()

    def read(self):
        """Get an iterator on all items from the queue.
        """
        reader = self.reader()
        while True:
            item = reader.get()
            if item is None:
                return
            yield item

    def reader(self):
        """Get a reader object you can use to read with a timeout.
        """
        return _PQ_Reader(self)


class ProgressStatus(object):
    def __init__(self, current, total=1.0):
        self.current = max(0.0, min(current, total))
        if total <= 0.0:
            self.total = 1.0
        else:
            self.total = total

    @property
    def ratio(self):
        return self.current / self.total

    @property
    def percentage(self):
        return '%d%%' % int(self.current / self.total)


def is_collection(dataset_path):
    with open(dataset_path) as fin:
        dataset_doc = json.load(fin)
        for data_resource in dataset_doc['dataResources']:
            if data_resource.get('isCollection', False):
                return True

    return False


def get_collection_type(dataset_path):
    with open(dataset_path) as fin:
        dataset_doc = json.load(fin)
        for data_resource in dataset_doc['dataResources']:
            if data_resource.get('isCollection', False):
                return data_resource['resType']

    return None


def need_denormalize(dataset_path):
    with open(dataset_path) as fin:
        dataset_doc = json.load(fin)
        if len(dataset_doc['dataResources']) > 1:
            return True
        return False


def get_dataset_sample(dataset, problem, dataset_sample_path=None):
    task_keywords = problem['problem']['task_keywords']
    sample_size = SAMPLE_SIZE

    if any(tk in [TaskKeyword.OBJECT_DETECTION, TaskKeyword.FORECASTING] for tk in task_keywords):
        logger.info('Not doing sampling for task %s', '_'.join([x.name for x in task_keywords]))
        return dataset

    try:
        target_name = problem['inputs'][0]['targets'][0]['column_name']
        for res_id in dataset:
            if ('https://metadata.datadrivendiscovery.org/types/DatasetEntryPoint'
                    in dataset.metadata.query([res_id])['semantic_types']):
                break
        else:
            res_id = next(iter(dataset))

        if any(tk in [TaskKeyword.VIDEO, TaskKeyword.IMAGE, TaskKeyword.AUDIO] for tk in task_keywords):
            sample_size = 100
        original_size = len(dataset[res_id])

        if hasattr(dataset[res_id], 'columns') and len(dataset[res_id]) > sample_size:
            labels = dataset[res_id].get(target_name)
            ratio = sample_size / original_size
            try:
                x_train, x_test, y_train, y_test = train_test_split(dataset[res_id], labels, random_state=RANDOM_SEED,
                                                                    test_size=ratio, stratify=labels)
            except:
                # Not using stratified sampling when the minority class has few instances, not enough for all the folds
                x_train, x_test, y_train, y_test = train_test_split(dataset[res_id], labels, random_state=RANDOM_SEED,
                                                                    test_size=ratio)
            dataset[res_id] = x_test
            logger.info('Sampling down data from %d to %d', original_size, len(dataset[res_id]))
            if dataset_sample_path:
                dataset.save(dataset_sample_path + 'datasetDoc.json')
                dataset_sample_uri = dataset_sample_path + 'datasetDoc.json'
                return dataset_sample_uri

        else:
            logger.info('Not doing sampling for small dataset (size = %d)', original_size)
    except:
        logger.error('Error sampling in dataset %s')

    return dataset


def get_internal_scoring_config(task_keywords):
    scoring_config = {'shuffle': 'true', 'method': 'K_FOLD',  'number_of_folds': '2',
                      'stratified': 'true' if TaskKeyword.CLASSIFICATION in task_keywords else 'false'
                      }

    return scoring_config


def create_outputfolders(folder_path):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)


def read_streams(process):
    error_msg = None
    try:
        _, stderr = process.communicate(timeout=5)
        error_msg = stderr.decode()
    except subprocess.TimeoutExpired:
        logger.error('Timeout error reading child process logs')

    return error_msg


def load_primitives_types():
    primitives_types = {}
    primitives = load_primitives_list()

    for primitive in primitives:
        primitives_types[primitive['python_path']] = primitive['type']

    return primitives_types



import os
import csv
import json
import sys
import time
import grpc
import logging
import subprocess
import pandas as pd
import d3m_automl_rpc.core_pb2_grpc as pb_core_grpc
from datetime import datetime
from os.path import join
from d3m.metadata.pipeline import Pipeline
from alphad3m.grpc_api.grpc_logger import LoggingStub
from d3m_automl_rpc.utils import encode_pipeline_description, decode_value
from d3m import index as d3m_index
from d3m.container import Dataset
from d3m.metadata.problem import Problem, PerformanceMetric
from d3m.utils import yaml_load_all, fix_uri
from alphad3m.grpc_api.grpc_client import do_search, do_score, do_train, do_test, do_export, do_describe, \
    do_load_solution, do_save_fitted_solution

import glob
import pkg_resources
import d3m.runtime
from d3m.metadata.pipeline import Pipeline

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')
logger = logging.getLogger(__name__)


D3MINPUTDIR = os.environ.get('D3MINPUTDIR')
D3MOUTPUTDIR = os.environ.get('D3MOUTPUTDIR')
D3MSTATICDIR = os.environ.get('D3MSTATICDIR')


def search_pipelines(datasets, time_bound=10, use_template=False):
    search_results_path = join(D3MOUTPUTDIR, 'temp', 'search_results.json')
    search_results = load_search_results(search_results_path)
    channel = grpc.insecure_channel('localhost:{0}'.format(os.environ.get('D3MPORT', 45042)))
    core = LoggingStub(pb_core_grpc.CoreStub(channel), logger)
    size = len(datasets)
    pipeline_template = None

    if use_template:
        pipeline_template = load_template()

    for i, dataset in enumerate(datasets):
        logger.info('Processing dataset "%s" (%d/%d)' % (dataset, i+1, size))
        start_time = datetime.now()

        dataset_train_path = join(D3MINPUTDIR, dataset, 'TRAIN/dataset_TRAIN/datasetDoc.json')
        problem_path = join(D3MINPUTDIR, dataset, 'TRAIN/problem_TRAIN/problemDoc.json')

        if not os.path.isfile(problem_path):
            logger.error('Problem file (%s) doesnt exist', problem_path)
            continue

        try:
            problem = Problem.load(problem_uri=fix_uri(problem_path))
        except:
            logger.exception('Error parsing problem')
            continue
        #debug_pipeline(fix_uri(dataset_train_path))
        task_keywords = '_'.join([x.name for x in problem['problem']['task_keywords']])
        search_id, pipelines = do_search(core, problem, dataset_train_path, time_bound=time_bound, pipelines_limit=0,
                                        pipeline_template=pipeline_template)
        #print(dataset, problem['problem']['performance_metrics'][0]['metric'].name, task_keywords)
        number_pipelines = len(pipelines)
        result = {'search_id': search_id, 'task': task_keywords, 'search_time': str(datetime.now() - start_time),
                  'pipelines': number_pipelines, 'best_time': 'None', 'best_score': 'None', 'all_scores': []}

        if number_pipelines > 0:
            best_time = sorted(pipelines.values(), key=lambda x: x[2])[0][2]
            sorted_pipelines = sorted(pipelines.items(), key=lambda x: x[1][0], reverse=True)
            all_scores = []

            for pipeline_id, (_, pipeline, pipeline_time) in sorted_pipelines:
                if use_template:  # FIXME: Pipeline's score is not calculate when working with fully defined template
                    pipeline_score = 1.0
                else:
                    pipeline_score = decode_value(pipeline[0].scores[0].value)['value']
                all_scores.append({'id': pipeline_id, 'score': pipeline_score, 'time': pipeline_time})
                #do_score(core, problem, [pipeline_id], dataset_train_path)
                #fitted_pipeline = do_train(core, [pipeline_id], dataset_train_path)
                #do_save_fitted_solution(core, fitted_pipeline)
                #do_test(core, fitted_pipeline, dataset_train_path.replace('TRAIN', 'TEST'))
                #do_export(core, fitted_pipeline)
                #do_describe(core, [pipeline_id])
                #pipeline_id =  do_load_solution(core, '/output/saved_pipeline')

            result['pipelines'] = number_pipelines
            result['best_time'] = best_time
            result['best_score'] = all_scores[0]['score']
            result['all_scores'] = all_scores

        search_results[dataset] = result

        with open(search_results_path, 'w') as fout:
            json.dump(search_results, fout, indent=4)


def evaluate_pipelines(datasets, top=10, option='fit-score'):
    statistics_path = join(D3MOUTPUTDIR, 'temp', 'statistics_datasets.csv')
    search_results_path = join(D3MOUTPUTDIR, 'temp', 'search_results.json')
    search_results = load_search_results(search_results_path)
    size = len(datasets)

    for i, dataset in enumerate(datasets):
        if dataset not in search_results:
            continue

        logger.info('Processing dataset "%s" (%d/%d)' % (dataset, i+1, size))
        dataset_train_path = join(D3MINPUTDIR, dataset, 'TRAIN/dataset_TRAIN/datasetDoc.json')
        dataset_test_path = join(D3MINPUTDIR, dataset, 'TEST/dataset_TEST/datasetDoc.json')
        dataset_score_path = join(D3MINPUTDIR, dataset, 'SCORE/dataset_SCORE/datasetDoc.json')
        problem_path = join(D3MINPUTDIR, dataset, 'TRAIN/problem_TRAIN/problemDoc.json')
        performance_top_pipelines = {}
        best_score = metric = best_id = 'None'

        for top_pipeline in search_results[dataset]['all_scores'][:top]:
            top_pipeline_id = top_pipeline['id']
            search_id = search_results[dataset]['search_id']
            logger.info('Scoring top pipeline id=%s' % top_pipeline_id)
            pipeline_path = join(D3MOUTPUTDIR, search_id, 'pipelines_searched', '%s.json' % top_pipeline_id)
            #pipeline_path = '/usr/src/app/resource/pipelines/example_metalearningdb.json'
            output_path = join(D3MOUTPUTDIR, 'temp', 'runtime_output', '%s_%s.csv' % (option, top_pipeline_id))
            try:
                command = create_command(option, top_pipeline_id, pipeline_path, output_path, problem_path,
                                         dataset_train_path, dataset_test_path, dataset_score_path)
                subprocess.call(command)
                if option == 'fit-score':
                    df = pd.read_csv(output_path)
                    score = round(df['value'][0], 6)
                    metric = df['metric'][0]
                    normalized_score = PerformanceMetric[metric].normalize(score)
                    performance_top_pipelines[top_pipeline_id] = (score, normalized_score)
                    logger.info('Scored top pipeline id=%s, %s=%.6f' % (top_pipeline_id, metric, score))
            except:
                logger.exception('Error during the process %s' % option)

        if len(performance_top_pipelines) > 0:
            best_pipeline = sorted(performance_top_pipelines.items(), key=lambda x: x[1][1], reverse=True)[0]
            best_id = best_pipeline[0]
            best_score = best_pipeline[1][0]
            logger.info('Best pipeline id=%s %s=%.6f' % (best_id, metric, best_score))

        row = [dataset, search_results[dataset]['pipelines'], search_results[dataset]['best_time'],
               search_results[dataset]['search_time'], best_score, metric, search_results[dataset]['task'], best_id]
        save_row(statistics_path, row)
        #create_dupms(search_id, performance_top_pipelines.keys())


def create_command(option, pipeline_id, pipeline_path, output_path, problem_path, train_path, test_path,
                   score_path=None, expose_outputs=False):
    command = [
        'python3', '-m', 'd3m',
        'runtime',
        '--volumes', D3MSTATICDIR,
        '--context', 'TESTING',
        '--random-seed', '0',
        option,
        '--pipeline', pipeline_path,
        '--problem', problem_path,
        '--input', train_path,
        '--test-input', test_path
        #'--output-run', join(D3MOUTPUTDIR, search_id, 'pipeline_runs', 'run_%s.yml' % top_pipeline_id)
    ]

    if expose_outputs:
        command += ['-E', join(D3MOUTPUTDIR, 'temp', 'runtime_output', pipeline_id)]
    if option == 'fit-produce':
        command += ['--output', output_path]
    elif option == 'fit-score':
        command += ['--score-input', score_path, '--scores', output_path]

    return command


def save_row(file_path, row):
    with open(file_path, 'a') as fout:
        writer = csv.writer(fout, delimiter='\t')
        writer.writerow(row)


def load_search_results(file_path):
    if not os.path.isfile(file_path):
        with open(file_path, 'w') as fout:
            json.dump({}, fout)

    with open(file_path) as fin:
        return json.load(fin)


def load_template():
    with open(join(os.path.dirname(__file__), '../resource/pipelines/example_metalearningdb.json')) as fin:
        json_pipeline = json.load(fin)

    d3m_pipeline = Pipeline.from_json_structure(json_pipeline, )
    grpc_pipeline = encode_pipeline_description(d3m_pipeline, ['RAW'], '/tmp')

    return grpc_pipeline


def create_dupms(search_id, top_pipelines):
    pipelines_list = []
    pipeline_runs_list = []

    for top_pipeline in top_pipelines:
        with open(join(D3MOUTPUTDIR, search_id, 'pipeline_runs', 'run_%s.yml' % top_pipeline)) as fin:
            for pipeline_run in yaml_load_all(fin):
                digest = pipeline_run['pipeline']['digest']
                pipeline_runs_list.append(json.dumps(pipeline_run))

        with open(join(D3MOUTPUTDIR, search_id, 'pipelines_searched', '%s.json' % top_pipeline)) as fin:
            pipeline = json.load(fin)
            pipeline['digest'] = digest
            pipeline['source'] = {'name': 'new_nyu'}
            pipelines_list.append(json.dumps(pipeline))

    with open(join(D3MOUTPUTDIR, 'temp', 'pipelines.json'), 'w') as fout:
        fout.write('\n'.join(pipelines_list))

    with open(join(D3MOUTPUTDIR, 'temp', 'pipeline_runs.json'), 'w') as fout:
        fout.write('\n'.join(pipeline_runs_list))


def create_inputs_pipelineprofiler(dataset):
    search_results_path = join(D3MOUTPUTDIR, 'temp', 'search_results.json')
    search_results = load_search_results(search_results_path)
    search_id = search_results[dataset]['search_id']
    pipelines = search_results[dataset]['all_scores']
    profiler_inputs = []

    for pipeline_info in pipelines:
        with open(join(D3MOUTPUTDIR, search_id, 'pipelines_searched/%s.json' % pipeline_info['id'])) as fin:
            json_pipeline = json.load(fin)

        score = float(pipeline_info['score'])
        problem = dataset
        start_time = datetime.utcnow().isoformat() + 'Z'
        metric = 'f1'
        end_time = datetime.utcnow().isoformat() + 'Z'
        normalized_score = PerformanceMetric[metric.upper()].normalize(score)
        pipeline_score = [{'metric': {'metric': metric}, 'value': score,
                           'normalized': normalized_score}]
        if 'digest' not in json_pipeline:
            json_pipeline['digest'] = json_pipeline['id']

        profiler_data = {
            'pipeline_id': json_pipeline['id'],
            'inputs': json_pipeline['inputs'],
            'steps': json_pipeline['steps'],
            'outputs': json_pipeline['outputs'],
            'pipeline_digest': json_pipeline['digest'],
            'problem': problem,
            'start': start_time,
            'end': end_time,
            'scores': pipeline_score,
            'pipeline_source': {'name': 'NYU'},
        }
        profiler_inputs.append(profiler_data)

        with open(join(D3MOUTPUTDIR, 'pipeline_profiler.json'), 'w') as fout:
            json.dump(profiler_inputs, fout)


def run_describe():
    pipeline_path = '/usr/src/app/resource/pipelines/example_metalearningdb.json'
    command = [
        'python3', '-m', 'd3m',
        'pipeline',
        'describe',
        pipeline_path,
    ]

    return_code = subprocess.call(command)
    logger.info('Describe pipeline process done, returned %d ' % return_code)


def debug_pipeline(dataset_uri):
    logger.info('Debugging pipeline step by step')
    start = time.time()
    dataset = Dataset.load(dataset_uri)
    duration = time.time() - start
    logger.info('Time after load dataset: %.5f' % duration)

    start = time.time()
    primitive_class = d3m_index.get_primitive('d3m.primitives.data_transformation.denormalize.Common')
    primitive_hyperparams = primitive_class.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    primitive_denormalize = primitive_class(hyperparams=primitive_hyperparams.defaults())
    primitive_output = primitive_denormalize.produce(inputs=dataset).value
    duration = time.time() - start
    logger.info('Time after denormoralize: %.5f' % duration)

    start = time.time()
    primitive_class = d3m_index.get_primitive('d3m.primitives.data_transformation.dataset_to_dataframe.Common')
    primitive_hyperparams = primitive_class.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    primitive_dataframe = primitive_class(hyperparams=primitive_hyperparams.defaults())
    primitive_output = primitive_dataframe.produce(inputs=primitive_output).value
    duration = time.time() - start
    logger.info('Time after dataset_to_dataframe: %.5f' % duration)

    start = time.time()
    primitive_class = d3m_index.get_primitive('d3m.primitives.data_transformation.column_parser.Common')
    primitive_hyperparams = primitive_class.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    primitive_dataframe = primitive_class(hyperparams=primitive_hyperparams.defaults())
    primitive_output = primitive_dataframe.produce(inputs=primitive_output).value
    duration = time.time() - start
    logger.info('Time after column_parser: %.5f' % duration)


def score_pipeline(task, json_pipeline):
    with pkg_resources.resource_stream('alphad3m', '../resource/pipelines/kfold_tabular_split.yaml') as fp:
        data_pipeline = Pipeline.from_yaml(fp)

    with pkg_resources.resource_stream('alphad3m', '../resource/pipelines/scoring.yaml') as fp:
        scoring_pipeline = Pipeline.from_yaml(fp)

    dataset_path = join(D3MINPUTDIR, task, 'openml_dataset_*/datasetDoc.json')
    dataset_path = glob.glob(dataset_path)[0]
    problem_path = join(D3MINPUTDIR, task, 'openml_problem_*/problemDoc.json')
    problem_path = glob.glob(problem_path)[0]

    dataset = Dataset.load(fix_uri(dataset_path))
    problem = Problem.load(problem_uri=fix_uri(problem_path))
    d3m_pipeline = Pipeline.from_json_structure(json_pipeline)
    metrics = problem['problem']['performance_metrics']
    scoring_config = {'shuffle': 'true', 'method': 'K_FOLD', 'number_of_folds': '10', 'stratified': 'true'}

    run_scores, run_results = d3m.runtime.evaluate(
        pipeline=d3m_pipeline,
        data_pipeline=data_pipeline,
        scoring_pipeline=scoring_pipeline,
        problem_description=problem,
        inputs=[dataset],
        data_params=scoring_config,
        metrics=metrics,
        volumes_dir=os.environ.get('D3MSTATICDIR', None),
        context=d3m.metadata.base.Context.TESTING,
        random_seed=0,
    )

    for result in run_results:
        if result.has_error():
            raise RuntimeError(result.pipeline_run.status['message'])

    try:
        scores = d3m.runtime.combine_folds([fold for fold in run_scores])
        logger.info('Cross-validation scores:\n%s'% scores.to_string())
        avg_score = round(scores['value'].mean(), 3)
    except:
        logger.error('Scoring pipeline')
        avg_score = 0

    logger.info('Task: %s, average score: %.3f' % (task, avg_score))

    return avg_score


def evaluate_openml_tasks():
    directory_path = join(D3MOUTPUTDIR, 'openml_pipelines')
    tasks = sorted([x for x in os.listdir(directory_path) if os.path.isdir(join(directory_path, x))])
    scores = {}
    for task in tasks:
        try:
            with open(join(directory_path, task, 'pipeline.json')) as fin:
                json_pipeline = json.load(fin)
        except:
            scores[task] = 0
            continue

        score = score_pipeline(task, json_pipeline)
        scores[task] = score

    for task, score in sorted(scores.items(), key=lambda x: x[0]):
        logger.info('Task: %s, average score: %.3f' % (task, score))


if __name__ == '__main__':
    datasets = sorted([x for x in os.listdir(D3MINPUTDIR) if os.path.isdir(join(D3MINPUTDIR, x))])
    if len(sys.argv) > 1:
        datasets = sys.argv[1:]
    search_pipelines(datasets, 5)
    evaluate_pipelines(datasets)



class NeuralNet():
    """
    This class specifies the base NeuralNet class. To define your own neural
    network, subclass this class and implement the functions below. The neural
    network does not consider the current player, and instead only deals with
    the canonical form of the board.

    See othello/NNet.py for an example implementation.
    """

    def __init__(self, game):
        pass

    def train(self, examples):
        """
        This function trains the neural network with examples obtained from
        self-play.

        Input:
            examples: a list of training examples, where each example is of form
                      (board, pi, v). pi is the MCTS informed policy vector for
                      the given board, and v is its value. The examples has
                      board in its canonical form.
        """
        pass

    def predict(self, board):
        """
        Input:
            board: current board in its canonical form.

        Returns:
            pi: a policy vector for the current board- a numpy array of length
                game.getActionSize
            v: a float in [-1,1] that gives the value of the current board
        """
        pass

    def save_checkpoint(self, folder, filename):
        """
        Saves the current neural network (with its parameters) in
        folder/filename
        """
        pass

    def load_checkpoint(self, folder, filename):
        """
        Loads parameters of the neural network from folder/filename
        """
        pass



"""Version of multiprocessing that doesn't use fork.

Necessary because of grpcio bug.
"""

import base64
import importlib
import logging
import multiprocessing.connection
import os
import pickle
import socket
from queue import Empty
import subprocess
import traceback
import sys


class Receiver(object):
    def __init__(self):
        self._listener = socket.socket(getattr(socket, 'AF_UNIX'))
        try:
            self._listener.setsockopt(socket.SOL_SOCKET,
                                      socket.SO_REUSEADDR, 1)
            address = multiprocessing.connection.arbitrary_address('AF_UNIX')
            self._listener.bind(address)
            self._listener.listen(1)
            self.address = self._listener.getsockname()
        except OSError:
            self._listener.close()
            raise

        self._sockets = [self._listener]

    def recv(self, timeout=None):
        done = False
        while not done:
            done = True
            for sock in multiprocessing.connection.wait(self._sockets,
                                                        timeout):
                if sock == self._listener:
                    s, addr = self._listener.accept()
                    conn = multiprocessing.connection.Connection(s.detach())
                    self._sockets.append(conn)
                    done = False
                else:
                    try:
                        msg = sock.recv()
                    except EOFError:
                        self._sockets.remove(sock)
                    else:
                        return msg
        raise Empty

    def send(self, msg):
        for sock in self._sockets:
            if sock != self._listener:
                sock.send(msg)

    def close(self):
        try:
            self._listener.close()
        finally:
            os.unlink(self.address)


def run_process(target, tag, msg_queue, **kwargs):
    """Call a Python function by name in a subprocess.

    :param target: Fully-qualified name of function to call.
    :param tag: Tag to add to logger to identify that process.
    :return: A `subprocess.Popen` object.
    """
    assert isinstance(msg_queue, Receiver)
    data = msg_queue.address, kwargs
    proc = subprocess.Popen(
        [
            sys.executable,
            '-c',
            'from alphad3m.multiprocessing import _invoke; _invoke(%r, %r)' % (
                tag, target
            ),
            base64.b64encode(pickle.dumps(data)),
        ],
        stdin=subprocess.PIPE, stderr=subprocess.PIPE)

    return proc


def _invoke(tag, target):
    """Invoked in the subprocess to setup logging and start the function.

    Arguments are read from ``sys.argv``.
    """
    data = pickle.loads(base64.b64decode(sys.argv[1]))
    address, kwargs = data

    tag = '{}-{}'.format(tag, os.getpid())

    logging.getLogger().handlers = []
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s:%(levelname)s:{}:%(name)s:%(message)s".format(tag),
        stream=sys.stdout)

    msg_queue = multiprocessing.connection.Client(address)

    module, function = target.rsplit('.', 1)
    module = importlib.import_module(module)
    function = getattr(module, function)

    try:
        function(msg_queue=msg_queue, **kwargs)
    except Exception:
        logging.exception("Uncaught exception in subprocess %s", tag)
        error = traceback.format_exc()
        sys.stderr.write(error)
        sys.exit(1)



import json
import shutil
import datamart_profiler
import pandas as pd
from datamart_materialize.d3m import D3mWriter
from os.path import join, exists


def create_d3mdataset(csv_path, destination_path, version='4.0.0'):
    metadata = datamart_profiler.process_dataset(csv_path)
    dataset_path = join(destination_path, 'datasetDoc.json')

    if exists(destination_path):
        shutil.rmtree(destination_path)

    writer = D3mWriter(
        dataset_id='internal_dataset',
        destination=destination_path,
        metadata=metadata,
        format_options={'need_d3mindex': True, 'version': version},
    )
    with open(csv_path, 'rb') as source:
        with writer.open_file('wb') as dest:
            shutil.copyfileobj(source, dest)
    writer.finish()

    return dataset_path


def create_d3mproblem(problem_config, csv_path, destination_path):
    if 'target_index' not in problem_config:
        target_index = pd.read_csv(csv_path).columns.get_loc(problem_config['target_name'])
        problem_config['target_index'] = target_index

    problem_path = join(destination_path, 'problemDoc.json')

    problem_json = {
          "about": {
            "problemID": "",
            "problemName": "",
            "problemDescription": "",
            "problemVersion": "4.0.0",
            "problemSchemaVersion": "4.0.0",
            "taskKeywords": problem_config.get('task_keywords', ["classification", "multiClass"])
          },
          "inputs": {
            "data": [
              {
                "datasetID": "internal_dataset",
                "targets": [
                  {
                    "targetIndex": 0,
                    "resID": "learningData",
                    "colIndex": problem_config.get('target_index'),
                    "colName": problem_config.get('target_name')
                  }
                ]
              }
            ],
            "performanceMetrics": [
              {
                "metric": problem_config.get('metric', "accuracy")
              }
            ]
          },
          "expectedOutputs": {
            "predictionsFile": "predictions.csv"
          }
        }

    with open(problem_path, 'w') as fout:
        json.dump(problem_json, fout, indent=4)

    return problem_path



import logging
import json
from os.path import join, dirname, isfile
from d3m import index
from collections import OrderedDict

logger = logging.getLogger(__name__)

PRIMITIVES_LIST_PATH = join(dirname(__file__), 'resource/primitives_list.json')
PRIMITIVES_HIERARCHY_PATH = join(dirname(__file__), 'resource/primitives_hierarchy.json')
INSTALLED_PRIMITIVES = sorted(index.search(), key=lambda x: x.endswith('SKlearn'), reverse=True)

BLACK_LIST = {
     # Not working primitives:
    'd3m.primitives.classification.random_classifier.Test',
    'd3m.primitives.classification.global_causal_discovery.ClassifierRPI',
    'd3m.primitives.classification.tree_augmented_naive_bayes.BayesianInfRPI',
    'd3m.primitives.classification.simple_cnaps.UBC',
    'd3m.primitives.classification.logistic_regression.UBC',
    'd3m.primitives.classification.multilayer_perceptron.UBC',
    'd3m.primitives.classification.canonical_correlation_forests.UBC',
    'd3m.primitives.regression.multilayer_perceptron.UBC',
    'd3m.primitives.regression.canonical_correlation_forests.UBC',
    'd3m.primitives.regression.linear_regression.UBC',
    'd3m.primitives.classification.inceptionV3_image_feature.Gator',
    'd3m.primitives.classification.search.Find_projections',
    'd3m.primitives.classification.search_hybrid.Find_projections',
    'd3m.primitives.regression.search_hybrid_numeric.Find_projections',
    'd3m.primitives.regression.search_numeric.Find_projections',
    'd3m.primitives.data_cleaning.binarizer.SKlearn',
    'd3m.primitives.feature_selection.rffeatures.Rffeatures',
    'd3m.primitives.feature_selection.mutual_info_classif.DistilMIRanking',
    'd3m.primitives.dimensionality_reduction.t_distributed_stochastic_neighbor_embedding.Tsne',
    'd3m.primitives.data_cleaning.string_imputer.SKlearn',
    'd3m.primitives.data_cleaning.tabular_extractor.Common',
    'd3m.primitives.data_cleaning.missing_indicator.SKlearn',
    'd3m.primitives.data_transformation.gaussian_random_projection.SKlearn',
    'd3m.primitives.data_transformation.sparse_random_projection.SKlearn',
    'd3m.primitives.feature_extraction.nk_sent2vec.Sent2Vec',
    'd3m.primitives.classification.mlp.BBNMLPClassifier',
    # Repeated primitives:
    'd3m.primitives.data_transformation.unary_encoder.DSBOX',
    'd3m.primitives.data_transformation.one_hot_encoder.TPOT',
    'd3m.primitives.data_transformation.one_hot_encoder.MakerCommon',
    'd3m.primitives.data_transformation.one_hot_encoder.PandasCommon',
    'd3m.primitives.feature_extraction.tfidf_vectorizer.BBNTfidfTransformer',
    'd3m.primitives.data_transformation.one_hot_encoder.DistilOneHotEncoder',
    'd3m.primitives.feature_selection.pca_features.Pcafeatures',
    'd3m.primitives.classification.random_forest.Common',
    # Poor performance:
    'd3m.primitives.classification.cover_tree.Fastlvm',
    'd3m.primitives.classification.linear_svc.DistilRankedLinearSVC',
    'd3m.primitives.classification.lstm.DSBOX',
    'd3m.primitives.regression.cover_tree.Fastlvm',
    'd3m.primitives.classification.bert_classifier.DistilBertPairClassification',
    'd3m.primitives.regression.global_causal_discovery.RegressorRPI',
    'd3m.primitives.regression.monomial.Test',
    'd3m.primitives.regression.rfm_precondition_ed_gaussian_krr.RFMPreconditionedGaussianKRR',
    'd3m.primitives.regression.rfm_precondition_ed_polynomial_krr.RFMPreconditionedPolynomialKRR',
    'd3m.primitives.regression.tensor_machines_regularized_least_squares.TensorMachinesRegularizedLeastSquares',
    'd3m.primitives.clustering.cluster_curve_fitting_kmeans.ClusterCurveFittingKMeans',
    'd3m.primitives.clustering.kmeans_clustering.UBC',
    'd3m.primitives.clustering.spectral_graph.SpectralClustering',
    'd3m.primitives.data_cleaning.greedy_imputation.DSBOX',
    'd3m.primitives.data_cleaning.iterative_regression_imputation.DSBOX',
    'd3m.primitives.data_cleaning.mean_imputation.DSBOX',
    'd3m.primitives.data_preprocessing.random_sampling_imputer.BYU',
    'd3m.primitives.data_transformation.imputer.DistilCategoricalImputer',
    'd3m.primitives.feature_extraction.feature_agglomeration.SKlearn',
    'd3m.primitives.feature_extraction.boc.UBC',
    'd3m.primitives.feature_extraction.bow.UBC',
    'd3m.primitives.natural_language_processing.glda.Fastlvm',
    'd3m.primitives.natural_language_processing.hdp.Fastlvm',
    'd3m.primitives.natural_language_processing.lda.Fastlvm',
    'd3m.primitives.classification.dummy.SKlearn',
    'd3m.primitives.regression.dummy.SKlearn',
    'd3m.primitives.data_cleaning.normalizer.SKlearn',
    'd3m.primitives.classification.ensemble_voting.DSBOX',
    # Long running times:
    'd3m.primitives.feature_selection.joint_mutual_information.AutoRPI',
    'd3m.primitives.feature_selection.score_based_markov_blanket.RPI',
    'd3m.primitives.feature_selection.simultaneous_markov_blanket.AutoRPI'
}


def get_primitive_class(primitive_name):
    return index.get_primitive(primitive_name)


def get_primitive_family(primitive_name):
    return get_primitive_class(primitive_name).metadata.to_json_structure()['primitive_family']


def get_primitive_algorithms(primitive_name):
    return get_primitive_class(primitive_name).metadata.to_json_structure()['algorithm_types']


def get_primitive_info(primitive_name):
    primitive_dict = get_primitive_class(primitive_name).metadata.to_json_structure()

    return {
            'id': primitive_dict['id'],
            'name': primitive_dict['name'],
            'version': primitive_dict['version'],
            'python_path': primitive_dict['python_path'],
            'digest': primitive_dict['digest'],
            'type': get_primitive_type(primitive_name)
    }


def get_primitive_type(primitive_name):
    primitive_type = get_primitive_family(primitive_name)
    #  Use the algorithm types as families because they are more descriptive
    if primitive_type in {'DATA_TRANSFORMATION', 'DATA_PREPROCESSING', 'DATA_CLEANING'}:
        algorithm_types = get_primitive_algorithms(primitive_name)
        primitive_type = algorithm_types[0]

    # Changing the primitive families using some predefined rules
    if primitive_name in {'d3m.primitives.data_cleaning.quantile_transformer.SKlearn',
                          'd3m.primitives.data_cleaning.normalizer.SKlearn',
                          'd3m.primitives.normalization.iqr_scaler.DSBOX'}:
        primitive_type = 'FEATURE_SCALING'

    elif primitive_name in {'d3m.primitives.feature_extraction.feature_agglomeration.SKlearn',
                            'd3m.primitives.feature_selection.mutual_info_classif.DistilMIRanking'}:
        primitive_type = 'FEATURE_SELECTION'

    elif primitive_name in {'d3m.primitives.feature_extraction.pca.SKlearn',
                            'd3m.primitives.feature_selection.pca_features.Pcafeatures',
                            'd3m.primitives.feature_extraction.truncated_svd.SKlearn',
                            'd3m.primitives.feature_extraction.pca_features.RandomizedPolyPCA',
                            'd3m.primitives.data_transformation.gaussian_random_projection.SKlearn',
                            'd3m.primitives.data_transformation.sparse_random_projection.SKlearn',
                            'd3m.primitives.data_transformation.fast_ica.SKlearn'}:
        primitive_type = 'DIMENSIONALITY_REDUCTION'  # Or should it be FEATURE_SELECTION ?

    elif primitive_name in {'d3m.primitives.classification.bert_classifier.DistilBertPairClassification',
                            'd3m.primitives.classification.text_classifier.DistilTextClassifier'}:
        primitive_type = 'TEXT_CLASSIFIER'

    elif primitive_name in {'d3m.primitives.data_transformation.enrich_dates.DistilEnrichDates'}:
        primitive_type = 'DATETIME_ENCODER'

    elif primitive_name in {'d3m.primitives.vertex_nomination.seeded_graph_matching.DistilVertexNomination',
                            'd3m.primitives.classification.gaussian_classification.JHU'}:
        primitive_type = 'VERTEX_CLASSIFICATION'

    elif primitive_name in {'d3m.primitives.graph_clustering.gaussian_clustering.JHU'}:
        primitive_type = 'COMMUNITY_DETECTION'

    elif primitive_name in {'d3m.primitives.data_transformation.load_single_graph.DistilSingleGraphLoader',
                            'd3m.primitives.data_transformation.load_graphs.DistilGraphLoader',
                            'd3m.primitives.data_transformation.load_graphs.JHU',
                            'd3m.primitives.data_transformation.load_edgelist.DistilEdgeListLoader'}:
        primitive_type = 'GRAPH_LOADER'

    elif primitive_name in {'d3m.primitives.feature_extraction.yolo.DSBOX'}:
        primitive_type = 'OBJECT_DETECTION'

    elif primitive_name in {'d3m.primitives.feature_construction.corex_text.DSBOX',
                            'd3m.primitives.data_transformation.encoder.DistilTextEncoder',
                            'd3m.primitives.feature_extraction.tfidf_vectorizer.SKlearn',
                            'd3m.primitives.feature_extraction.count_vectorizer.SKlearn',
                            'd3m.primitives.feature_extraction.boc.UBC',
                            'd3m.primitives.feature_extraction.bow.UBC',
                            'd3m.primitives.feature_extraction.nk_sent2vec.Sent2Vec',
                            'd3m.primitives.feature_extraction.tfidf_vectorizer.BBNTfidfTransformer'}:
        primitive_type = 'TEXT_FEATURIZER'

    elif primitive_name in {'d3m.primitives.feature_extraction.image_transfer.DistilImageTransfer',
                            'd3m.primitives.feature_extraction.resnet50_image_feature.DSBOX'}:
        primitive_type = 'IMAGE_FEATURIZER'

    elif primitive_name in {'d3m.primitives.feature_extraction.audio_transfer.DistilAudioTransfer'}:
        primitive_type = 'AUDIO_FEATURIZER'

    elif primitive_name in {'d3m.primitives.data_transformation.audio_reader.DistilAudioDatasetLoader'}:
        primitive_type = 'AUDIO_READER'

    elif primitive_name in {'d3m.primitives.feature_extraction.resnext101_kinetics_video_features.VideoFeaturizer'}:
        primitive_type = 'VIDEO_FEATURIZER'

    elif primitive_name in {'d3m.primitives.feature_extraction.random_projection_timeseries_featurization.DSBOX'}:
        primitive_type = 'TIME_SERIES_FEATURIZER'

    elif primitive_name in {'d3m.primitives.data_transformation.time_series_to_list.DSBOX'}:
        primitive_type = 'TIME_SERIES_READER'

    elif primitive_name in {'d3m.primitives.data_transformation.grouping_field_compose.Common'}:
        primitive_type = 'TIME_SERIES_GROUPER'

    elif primitive_name in {'d3m.primitives.data_transformation.dataframe_to_tensor.DSBOX'}:
        primitive_type = 'TO_TENSOR'

    if primitive_type == 'ENCODE_ONE_HOT':
        primitive_type = 'CATEGORICAL_ENCODER'

    return primitive_type


def load_primitives_hierarchy():
    if isfile(PRIMITIVES_HIERARCHY_PATH):
        with open(PRIMITIVES_HIERARCHY_PATH) as fin:
            primitives = json.load(fin)
        logger.info('Loading primitives info from file')

        return primitives

    primitives = {}
    for primitive_name in INSTALLED_PRIMITIVES:
        if primitive_name not in BLACK_LIST:
            try:
                primitive_type = get_primitive_type(primitive_name)
            except Exception as e:
                logger.error('Loading metadata about primitive %s', primitive_name, exc_info=e)
                continue

            if primitive_type not in primitives:
                primitives[primitive_type] = []
            primitives[primitive_type].append(primitive_name)

    with open(PRIMITIVES_HIERARCHY_PATH, 'w') as fout:
        json.dump(OrderedDict(sorted(primitives.items())), fout, indent=4)
    logger.info('Loading primitives info from D3M index')

    return primitives


def load_primitives_list():
    if isfile(PRIMITIVES_LIST_PATH):
        with open(PRIMITIVES_LIST_PATH) as fin:
            primitives = json.load(fin)
        logger.info('Loading primitives info from file')

        return primitives

    primitives = []
    for primitive_name in INSTALLED_PRIMITIVES:
            try:
                primitive_info = get_primitive_info(primitive_name)
            except Exception as e:
                logger.error('Loading metadata about primitive %s', primitive_name, exc_info=e)
                continue
            primitives.append(primitive_info)

    with open(PRIMITIVES_LIST_PATH, 'w') as fout:
        json.dump(primitives, fout, indent=4)
    logger.info('Loading primitives info from D3M index')

    return primitives



import re
import json
import logging
import math
import hashlib
import datamart_profiler
import pandas as pd
from os.path import join, dirname, exists
from metalearn import Metafeatures
from alphad3m.metalearning.database import load_metalearningdb
from sklearn.metrics.pairwise import cosine_similarity
from d3m.metadata.problem import TaskKeywordBase

logger = logging.getLogger(__name__)

DATASETS_FOLDER_PATH = '/Users/rlopez/D3M/datasets/seed_datasets_current/'
PRECALCULATED_METAFEATURES_PATH = join(dirname(__file__), '../resource/precalculated_metafeatures.json')
PRECALCULATED_DATAPROFILES_PATH = join(dirname(__file__), '../resource/precalculated_dataprofiles.json')
PRECALCULATED_TASKKEYWORDS_PATH = join(dirname(__file__), '../resource/precalculated_taskkeywords.json')


def load_precalculated_data(mode):
    if mode == 'metafeatures':
        file_path = PRECALCULATED_METAFEATURES_PATH
    elif mode == 'dataprofiles':
        file_path = PRECALCULATED_DATAPROFILES_PATH
    elif mode == 'task_keywords':
        file_path = PRECALCULATED_TASKKEYWORDS_PATH
    else:
        raise ValueError('Unknown mode "%s" to load data' % mode)

    if exists(file_path):
        with open(file_path) as fin:
            return json.load(fin)

    return {}


def get_unique_datasets():
    pipeline_runs = load_metalearningdb()
    datasets = set()

    for pipeline_run in pipeline_runs:
        problem_id = pipeline_run['problem']['id']
        dataset_id = get_dataset_id(problem_id)
        datasets.add(dataset_id)

    logger.info('Found %d unique datasets', len(datasets))

    return sorted(datasets)


def get_dataset_id(problem_id):
    # Remove suffixes 'TRAIN' and 'problem' from the dataset name
    dataset_id = re.sub('_TRAIN$', '', problem_id)
    dataset_id = re.sub('_problem$', '', dataset_id)

    return dataset_id


def load_task_info(dataset_id):
    possible_names = [join(DATASETS_FOLDER_PATH, dataset_id), join(DATASETS_FOLDER_PATH, dataset_id + '_MIN_METADATA'),
                      join(DATASETS_FOLDER_PATH, dataset_id.replace('_MIN_METADATA', ''))]
    # All possible names of the datasets on disk, with/without the suffix 'MIN_METADATA'

    for dataset_folder_path in possible_names:
        if exists(dataset_folder_path):
            break
    else:
        raise FileNotFoundError('Dataset %s not found' % dataset_id)

    dataset_path = join(dataset_folder_path, 'TRAIN/dataset_TRAIN/tables/learningData.csv')
    problem_path = join(dataset_folder_path, 'TRAIN/problem_TRAIN/problemDoc.json')

    with open(problem_path) as fin:
        problem_doc = json.load(fin)
        task_keywords = problem_doc['about']['taskKeywords']
        target_column = problem_doc['inputs']['data'][0]['targets'][0]['colName']

    return dataset_path, target_column, task_keywords


def extract_metafeatures(dataset_path, target_column):
    data = pd.read_csv(dataset_path)
    Y = data[target_column]
    Y = pd.Series([str(i) for i in Y], name=target_column)  # Cast to string to get metalearn lib working correctly
    X = data.drop(columns=[target_column])
    metafeatures = Metafeatures()
    mfs = metafeatures.compute(X, Y, seed=0, timeout=300)

    return mfs


def extract_dataprofiles(dataset_path, target_column, ignore_target_column=False):
    metadata = datamart_profiler.process_dataset(dataset_path, coverage=False)
    feature_types = set()
    missing_values = False
    for item in metadata['columns']:
        if ignore_target_column and item['name'] == target_column:
            continue
        identified_types = item['semantic_types'] if len(item['semantic_types']) > 0 else [item['structural_type']]
        for feature_type in identified_types:
            feature_types.add(feature_type)

        if 'missing_values_ratio' in item and item['name'] != target_column:
            missing_values = True

    dps = {'feature_types': sorted(feature_types), 'missing_values': missing_values}

    return dps


def extract_metafeatures_mldb():
    datasets = get_unique_datasets()
    metafeatures = load_precalculated_data('metafeatures')

    for dataset_id in datasets:
        logger.info('Calculating metafeatures for dataset %s...', dataset_id)
        if dataset_id not in metafeatures:
            try:
                dataset_path, target_column, _ = load_task_info(dataset_id)
                mfs = extract_metafeatures(dataset_path, target_column)
                metafeatures[dataset_id] = mfs
                logger.info('Metafeatures successfully calculated for dataset %s', dataset_id)
                with open(PRECALCULATED_METAFEATURES_PATH, 'w') as fout:
                    json.dump(metafeatures, fout, indent=4, sort_keys=True)
            except Exception as e:
                logger.error(str(e))
        else:
            logger.info('Using pre-calculated metafeatures for dataset %s', dataset_id)

    return metafeatures


def extract_dataprofiles_mldb():
    datasets = get_unique_datasets()
    dataprofiles = load_precalculated_data('dataprofiles')

    for dataset_id in datasets:
        logger.info('Calculating data profiles for dataset %s...', dataset_id)
        if dataset_id not in dataprofiles:
            try:
                dataset_path, target_column, _ = load_task_info(dataset_id)
                dps = extract_dataprofiles(dataset_path, target_column)
                dataprofiles[dataset_id] = dps
                logger.info('Data profiles successfully calculated for dataset %s', dataset_id)
                with open(PRECALCULATED_DATAPROFILES_PATH, 'w') as fout:
                    json.dump(dataprofiles, fout, indent=4, sort_keys=True)
            except Exception as e:
                logger.error(str(e))
        else:
            logger.info('Using pre-calculated data profiles for dataset %s', dataset_id)

    return dataprofiles


def extract_taskkeywords_mldb():
    datasets = get_unique_datasets()
    task_keywords = load_precalculated_data('task_keywords')

    for dataset_id in datasets:
        logger.info('Calculating task keywords for dataset %s...', dataset_id)
        if dataset_id not in task_keywords:
            try:
                _, _, keywords = load_task_info(dataset_id)
                task_keywords[dataset_id] = {'task_keywords': keywords}
                logger.info('Task keywords successfully calculated for dataset %s', dataset_id)
                with open(PRECALCULATED_TASKKEYWORDS_PATH, 'w') as fout:
                    json.dump(task_keywords, fout, indent=4, sort_keys=True)
            except Exception as e:
                logger.error(str(e))
        else:
            logger.info('Using pre-calculated task keywords for dataset %s', dataset_id)

    return task_keywords


def create_metafeatures_vector(metafeatures, metafeature_indices):
    vector = []

    for metafeatures_id in metafeature_indices:
        value = metafeatures[metafeatures_id]['value']
        if isinstance(value, str):
            value = int(hashlib.sha256(value.encode('utf-8')).hexdigest(), 16) % 256
        elif math.isnan(value) or math.isinf(value):
            value = 0
        vector.append(value)

    return vector


def create_dataprofiles_vector(dataprofiles, dataprofile_indices):
    vector = []

    for dataprofile_id in dataprofile_indices:
        if dataprofile_id in dataprofiles['feature_types']:
            value = 1
        else:
            value = 0
        vector.append(value)

    value = 1 if dataprofiles['missing_values'] else 0
    vector.append(value)  # Add an extra value corresponding to the missing values data

    return vector


def create_taskkeywords_vector(task_keywords, taskkeyword_indices):
    vector = []

    for taskkeyword_id in taskkeyword_indices:
        if taskkeyword_id in task_keywords['task_keywords']:
            value = 1
        else:
            value = 0
        vector.append(value)

    return vector


def create_metafeatures_vectors_mldb(metafeature_indices):
    vectors = {}
    metafeature_datasets = load_precalculated_data('metafeatures')

    for id_dataset, metafeatures in metafeature_datasets.items():
        vector = create_metafeatures_vector(metafeatures, metafeature_indices)
        vectors[id_dataset] = vector

    return vectors


def create_dataprofiles_vectors_mldb(dataprofile_indices):
    vectors = {}
    dataprofile_datasets = load_precalculated_data('dataprofiles')

    for id_dataset, dataprofiles in dataprofile_datasets.items():
        vector = create_dataprofiles_vector(dataprofiles, dataprofile_indices)
        vectors[id_dataset] = vector

    return vectors


def create_taskkeywords_vectors_mldb(taskkeyword_indices):
    vectors = {}
    taskkeyword_datasets = load_precalculated_data('task_keywords')

    for id_dataset, task_keywords in taskkeyword_datasets.items():
        vector = create_taskkeywords_vector(task_keywords, taskkeyword_indices)
        vectors[id_dataset] = vector

    return vectors


def load_metafeatures_vectors(dataset_path, target_column):
    mfs = extract_metafeatures(dataset_path, target_column)
    metafeature_indices = Metafeatures.list_metafeatures(group='all')
    target_metafeatures_vector = create_metafeatures_vector(mfs, metafeature_indices)
    metalearningdb_vectors = create_metafeatures_vectors_mldb(metafeature_indices)

    return metalearningdb_vectors, target_metafeatures_vector


def load_profiles_vectors(dataset_path, target_column):
    dps = extract_dataprofiles(dataset_path, target_column)
    dataprofile_indices = [v for k, v in datamart_profiler.types.__dict__.items() if not k.startswith('_')]
    target_dataprofile_vector = create_dataprofiles_vector(dps, dataprofile_indices)
    metalearningdb_vectors = create_dataprofiles_vectors_mldb(dataprofile_indices)

    return metalearningdb_vectors, target_dataprofile_vector


def load_taskkeyword_vectors(task_keywords):
    taskkeyword_indices = sorted([keyword for keyword in TaskKeywordBase.get_map().keys() if keyword is not None])
    target_dataprofile_vector = create_taskkeywords_vector({'task_keywords': task_keywords}, taskkeyword_indices)
    metalearningdb_vectors = create_taskkeywords_vectors_mldb(taskkeyword_indices)

    return metalearningdb_vectors, target_dataprofile_vector


def calculate_similarity(metalearningdb_vectors, target_vector, threshold):
    similar_datasets = {}
    for id_dataset, vector in metalearningdb_vectors.items():
        similarity = cosine_similarity([target_vector], [vector]).flat[0]
        if similarity > threshold:
            similar_datasets[id_dataset] = round(similarity, 5)

    return similar_datasets


def similarity_repr(dataset_similarities):
    similarity_string = []

    for dataset_similarity in sorted(dataset_similarities.items(), key=lambda x: x[1], reverse=True):
        pretty_string = '%s=%.2f' % dataset_similarity
        similarity_string.append(pretty_string)

    return ', '.join(similarity_string)


def get_similar_datasets(mode, dataset_path, target_column, task_keywords, threshold=0.8, combined=False):
    vectors_taskkeywords, target_vector_taskkeywords = load_taskkeyword_vectors(task_keywords)

    if mode == 'metafeatures':
        vectors_dataset, target_vector_dataset = load_metafeatures_vectors(dataset_path, target_column)
    elif mode == 'dataprofiles':
        vectors_dataset, target_vector_dataset = load_profiles_vectors(dataset_path, target_column)
    else:
        raise ValueError('Unknown mode "%s" to load data' % mode)

    if combined:
        # Concatenate the vectors of the dataset and task keywords
        for id_dataset in vectors_dataset:
            vectors_dataset[id_dataset] += vectors_taskkeywords[id_dataset]
        target_vector_dataset += target_vector_taskkeywords
        similar_datasets = calculate_similarity(vectors_dataset, target_vector_dataset, threshold)
        logger.info('Similar datasets found using both information:\n%s', similarity_repr(similar_datasets))
    else:
        similar_datasets = calculate_similarity(vectors_taskkeywords, target_vector_taskkeywords, threshold)
        logger.info('Similar datasets found using task_keywords features:\n%s', similarity_repr(similar_datasets))
        vectors_dataset = {k: vectors_dataset[k] for k in similar_datasets}  # Use only the similar datasets
        similar_datasets = calculate_similarity(vectors_dataset, target_vector_dataset, threshold)
        logger.info('Similar datasets found using %s features:\n%s', mode, similarity_repr(similar_datasets))

    return similar_datasets



import logging
import numpy as np
from smac.configspace import ConfigurationSpace
from ConfigSpace.hyperparameters import IntegerHyperparameter, FloatHyperparameter, CategoricalHyperparameter, \
    OrdinalHyperparameter
from smac.facade.smac_ac_facade import SMAC4AC
from smac.scenario.scenario import Scenario
from alphad3m.hyperparameter_tuning.primitive_config import load_primitive_configspace, load_hyperparameters


MAX_RUNS = 100
logger = logging.getLogger(__name__)


def build_configspace(primitives):
    # Build Configuration Space which defines all parameters and their ranges
    configspace = ConfigurationSpace()
    for primitive in primitives:
        load_primitive_configspace(configspace, primitive)

    return configspace


def get_new_hyperparameters(primitive_name, configspace):
    hyperparameters = load_hyperparameters(primitive_name)
    new_hyperparameters = {}

    for hyperparameter_name in hyperparameters:
        hyperparameter_config_name = primitive_name + '|' + hyperparameter_name
        hyperparameter_config_name_case = hyperparameter_config_name + '|case'
        if hyperparameter_config_name in configspace:
            value = None if configspace[hyperparameter_config_name] == 'None' \
                else configspace[hyperparameter_config_name]
            new_hyperparameters[hyperparameter_name] = value
            logger.info('New value for %s=%s', hyperparameter_config_name, new_hyperparameters[hyperparameter_name])
        elif hyperparameter_config_name_case in configspace:
            case = configspace[hyperparameter_config_name_case]
            value = None if configspace[hyperparameter_config_name + '|' + case] == 'None' \
                else configspace[hyperparameter_config_name + '|' + case]
            new_hyperparameters[hyperparameter_name] = {'case': case,
                                                        'value': value}
            logger.info('New value for %s=%s', hyperparameter_config_name, new_hyperparameters[hyperparameter_name])

    return new_hyperparameters


class HyperparameterTuning(object):
    def __init__(self, primitives):
        self.configspace = build_configspace(primitives)
        # Avoiding too many iterations
        self.runcount = 1

        for param in self.configspace.get_hyperparameters():
            if isinstance(param, IntegerHyperparameter):
                self.runcount *= (param.upper - param.lower)
            elif isinstance(param, CategoricalHyperparameter):
                self.runcount *= len(param.choices)
            elif isinstance(param, OrdinalHyperparameter):
                self.runcount *= len(param.sequence)
            elif isinstance(param, FloatHyperparameter):
                self.runcount = MAX_RUNS
                break

        self.runcount = min(self.runcount, MAX_RUNS)

    def tune(self, runner, wallclock, output_dir):
        # Scenario object
        cutoff = wallclock / (self.runcount / 10)  # Allow long pipelines to try to execute one fourth of the iterations limit
        scenario = Scenario({'run_obj': 'quality',  # We optimize quality (alternatively runtime)
                             'runcount-limit': self.runcount,  # Maximum function evaluations
                             'wallclock-limit': wallclock,
                             'cutoff_time': cutoff,
                             'cs': self.configspace,  # Configuration space
                             'deterministic': 'true',
                             'output_dir': output_dir,
                             'abort_on_first_run_crash': False
                             })
        smac = SMAC4AC(scenario=scenario, rng=np.random.RandomState(0), tae_runner=runner)
        best_configuration = smac.optimize()
        min_cost = float('inf')
        best_scores = {}

        for _, run_data in smac.get_runhistory().data.items():
            if run_data.cost < min_cost:
                min_cost = run_data.cost
                best_scores = run_data.additional_info

        return best_configuration, best_scores



"""The AutoML class, that creates pipelines, train, and run them.

We use multiprocessing to run training in separate processes, sending messages
back to this process via a Queue.
"""

from concurrent import futures
import datetime
import grpc
import json
import logging
import os
import pickle
from queue import Empty, Queue
from sqlalchemy import select
from sqlalchemy.orm import aliased, joinedload, lazyload
from sqlalchemy.sql import func
from os.path import join, exists
import shutil
import multiprocessing
import subprocess
import threading
import time
import d3m_automl_rpc.core_pb2_grpc as pb_core_grpc
from uuid import uuid4, UUID
from alphad3m import __version__
from alphad3m.multiprocessing import Receiver, run_process
from alphad3m.grpc_api import grpc_server
from alphad3m.data_ingestion.data_profiler import profile_data
from alphad3m.pipeline_synthesis.templates import generate_pipelines
from alphad3m.utils import Observable, ProgressStatus, is_collection, get_dataset_sample, create_outputfolders, \
    read_streams, get_internal_scoring_config
from alphad3m.schema import database
from alphad3m.schema.convert import to_d3m_json
from alphad3m.data_ingestion.data_reader import create_d3mdataset, create_d3mproblem
from d3m.container import Dataset
from d3m.metadata.problem import TaskKeyword, parse_problem_description
from d3m.metadata import pipeline as pipeline_module


PIPELINES_TO_TUNE = 0  # Number of pipelines (top k) to be tuned.
TIME_TO_TUNE = 0.15  # The ratio of the time to be used for the tuning phase.
TIME_TO_SCORE = 5  # In minutes. Internal time to score a pipeline during the searching phase.
MAX_RUNNING_TIME = 43800  # In minutes. If time is not provided for the searching either scoring, run it for 1 month.
MAX_RUNNING_PROCESSES = int(os.environ.get('D3MCPU', multiprocessing.cpu_count()))  # Number of processes to be used
SEARCH_STRATEGIES = ['TEMPLATES', 'ALPHA_AUTOML']

logger = logging.getLogger(__name__)


class AutoML(Observable):
    def __init__(self, output_folder):
        Observable.__init__(self)
        self.output_folder = output_folder
        #  Internal folders
        self.internal_folder = os.path.join(self.output_folder, 'temp')
        self.runtime_folder = os.path.join(self.output_folder, 'temp', 'runtime_output')
        # TODO: Remove the runtime_folder. Use internal_folder instead
        create_outputfolders(self.internal_folder)
        create_outputfolders(self.runtime_folder)
        self.db_filename = os.path.join(self.internal_folder, 'db.sqlite3')

        logger.info("output_folder=%r", self.output_folder)

        self.dbengine, self.DBSession = database.connect(self.db_filename)

        self.sessions = {}
        self.executor = ThreadPoolExecutor(max_workers=16)
        self._run_queue = Queue()
        self._run_thread = threading.Thread(target=self._pipeline_running_thread)
        self._run_thread.setDaemon(True)
        self._run_thread.start()

        logger.warning("AlphaD3M started, version=%s", __version__)

    def run_search(self, dataset_path, problem_path=None, problem_config=None, timeout=None):
        if not dataset_path.endswith('datasetDoc.json'):  # Convert to D3M format
            logger.info('Reiceving a raw dataset, converting to D3M format')
            dataset_folder = os.path.join(self.output_folder, 'temp', 'dataset_d3mformat', 'dataset')
            problem_folder = os.path.join(self.output_folder, 'temp', 'dataset_d3mformat', 'problem')
            create_outputfolders(problem_folder)
            problem_path = create_d3mproblem(problem_config, dataset_path, problem_folder)
            dataset_path = create_d3mdataset(dataset_path, dataset_folder)

        if dataset_path[0] == '/':
            dataset_path = 'file://' + dataset_path

        problem = parse_problem_description(problem_path)
        task_keywords = problem['problem']['task_keywords']
        # Create session
        session = Session(self, problem, self.output_folder, self.DBSession)
        logger.info('Dataset: %s, task: %s, metrics: %s', dataset_path, '_'.join([x.name for x in task_keywords]),
                    session.metrics)
        self.sessions[session.id] = session

        # Create pipelines, NO TUNING
        with session.with_observer_queue() as queue:
            self.build_pipelines(session.id, dataset_path, task_keywords, session.metrics, timeout, None, None,
                                 tune=PIPELINES_TO_TUNE)

            while queue.get(True)[0] != 'done_searching':
                pass

        '''logger.info('Tuning pipelines...')

        # Now do tuning, when we already have written out some solutions
        with session.with_observer_queue() as queue:
            session.tune_when_ready()
            while queue.get(True)[0] != 'done_searching':
                pass'''

    def run_server(self, port=None):
        """Spin up the gRPC server to receive requests from a TA3 system.

        This is called by the ``ta2_serve`` executable. It is part of the
        TA2+TA3 evaluation.
        """

        if not port:
            port = 45042
        core_rpc = grpc_server.CoreService(self)
        server = grpc.server(self.executor, options=(("grpc.so_reuseport", 0),))
        pb_core_grpc.add_CoreServicer_to_server(
            core_rpc, server)
        server.add_insecure_port('[::]:%d' % port)
        logger.info("Started gRPC server on port %d", port)
        server.start()
        while True:
            time.sleep(60)

    def new_session(self, problem):
        session = Session(self, problem, self.output_folder, self.DBSession)
        self.sessions[session.id] = session

        return session.id

    def finish_session(self, session_id):
        session = self.sessions.pop(session_id)
        session.close()

    def stop_session(self, session_id):
        session = self.sessions[session_id]
        session.stop_requested = True

    def get_pipeline(self, pipeline_id):
        db = self.DBSession()
        try:
            pipeline = (
                db.query(database.Pipeline)
                    .filter(database.Pipeline.id == pipeline_id)
                    .options(joinedload(database.Pipeline.modules),
                             joinedload(database.Pipeline.connections))
            ).one()

            json_pipeline = to_d3m_json(pipeline)
            d3m_pipeline = pipeline_module.Pipeline.from_json_structure(json_pipeline, )
            # FIXME: from_json_structure is returning a different digest

            return d3m_pipeline

        finally:
            db.close()

    def get_pipeline_scores(self, pipeline_id):
        db = self.DBSession()
        try:
            # Find most recent evaluation
            evaluation_id = (
                select([database.Evaluation.id])
                .where(database.Evaluation.pipeline_id == pipeline_id)
                .order_by(database.Evaluation.date.desc())
            ).as_scalar()
            # Get scores from that evaluation
            scores = db.query(
                select([func.avg(database.EvaluationScore.value),
                        database.EvaluationScore.metric])
                .where(
                    database.EvaluationScore.evaluation_id ==
                    evaluation_id
                )
                .group_by(database.EvaluationScore.metric)
            ).all()
            return {metric: value for value, metric in scores}
        finally:
            db.close()

    def score_pipeline(self, pipeline_id, metrics, dataset_uri, problem, scoring_config, timeout_run):
        if timeout_run is None:
            timeout_run = MAX_RUNNING_TIME

        timeout_run = timeout_run * 60  # Minutes to seconds

        job = ScoreJob(self, pipeline_id, dataset_uri, metrics, problem, scoring_config, timeout_run)
        self._run_queue.put(job)
        return id(job)

    def train_pipeline(self, pipeline_id, dataset, problem, steps_to_expose):
        job = TrainJob(self, pipeline_id, dataset, problem, steps_to_expose)
        self._run_queue.put(job)
        return id(job)

    def test_pipeline(self, pipeline_id, dataset, steps_to_expose):
        job = TestJob(self, pipeline_id, dataset, steps_to_expose)
        self._run_queue.put(job)
        return id(job)

    def save_pipeline(self, session_id, pipeline_id, problem):
        pipeline_id = str(pipeline_id)
        session_id = str(session_id)
        pipeline_path = join(self.output_folder, session_id, 'pipelines_scored', '%s.json' % pipeline_id)

        if exists(pipeline_path):
            pipeline_folder_path = join(self.output_folder, 'temp', 'saved_pipelines', pipeline_id)
            create_outputfolders(pipeline_folder_path)
            shutil.copyfile(pipeline_path, join(pipeline_folder_path, 'pipeline.json'))
            problem_json = problem.to_json_structure()

            with open(join(pipeline_folder_path, 'problem.json'), 'w') as fout:
                json.dump(problem_json, fout, indent=2)

            return 'file://' + pipeline_folder_path

        return None

    def load_pipeline(self, session_id, pipeline):
        session = self.sessions[session_id]
        db = self.DBSession()
        dataset = 'N/A'

        pipeline_database = database.Pipeline(origin='Loaded pipeline', dataset=dataset)

        # TODO: Do it on d3mpipeline_generator.py
        def make_module(package, version, name):
            pipeline_module = database.PipelineModule(
                pipeline=pipeline_database,
                package=package, version=version, name=name)
            db.add(pipeline_module)
            return pipeline_module

        def make_data_module(name):
            return make_module('data', '0.0', name)

        def make_primitive_module(name):
            if name[0] == '.':
                name = 'd3m.primitives' + name
            return make_module('d3m', '2018.7.10', name)

        def connect(from_module, to_module,
                    from_output='produce', to_input='inputs'):
            db.add(database.PipelineConnection(pipeline=pipeline_database,
                                               from_module=from_module,
                                               to_module=to_module,
                                               from_output_name=from_output,
                                               to_input_name=to_input))

        def set_hyperparams(module, **hyperparams):
            db.add(database.PipelineParameter(
                pipeline=pipeline_database, module=module,
                name='hyperparams', value=pickle.dumps(hyperparams),
            ))

        try:
            # TODO: Use pipeline input for this
            if dataset:
                input_data = make_data_module('dataset')
                db.add(database.PipelineParameter(
                    pipeline=pipeline_database, module=input_data,
                    name='targets', value=pickle.dumps('targets'),
                ))
                db.add(database.PipelineParameter(
                    pipeline=pipeline_database, module=input_data,
                    name='features', value=pickle.dumps('features'),
                ))
            prev_step = None
            prev_steps = {}
            count_template_steps = 0
            for pipeline_step in pipeline['steps']:
                if pipeline_step['type'] == 'PRIMITIVE':
                    step = make_primitive_module(pipeline_step['primitive']['python_path'])
                    if 'outputs' in pipeline_step:
                        for output in pipeline_step['outputs']:
                            prev_steps['steps.%d.%s' % (count_template_steps, output['id'])] = step

                    count_template_steps += 1
                    if 'hyperparams' in pipeline_step:
                        hyperparams = {}
                        for hyper, desc in pipeline_step['hyperparams'].items():
                            hyperparams[hyper] = {'type': desc['type'], 'data': desc['data']}
                        set_hyperparams(step, **hyperparams)
                else:
                    # TODO In the future we should be able to handle subpipelines
                    break
                if prev_step:
                    if 'arguments' in pipeline_step:
                        for argument, desc in pipeline_step['arguments'].items():
                            connect(prev_steps[desc['data']], step, from_output=desc['data'].split('.')[-1],
                                    to_input=argument)
                    connect(prev_step, step, from_output='index', to_input='index')
                else:
                    connect(input_data, step, from_output='dataset')
                prev_step = step
            db.add(pipeline_database)
            db.commit()
            pipeline_id = pipeline_database.id
            logger.info("Created fixed pipeline %s", pipeline_id)
            session.write_searched_pipeline(pipeline_id)

            session.notify('new_fixed_pipeline', pipeline_id=pipeline_id)
            with session.lock:
                session.pipelines.add(pipeline_id)
                # Force working=True so we get 'done_searching' even if no pipeline
                # gets created
                session.working = False
        finally:
            db.close()

        return pipeline_id

    def save_fitted_pipeline(self, pipeline_id):
        fitted_pipeline_path = join(self.runtime_folder, 'fitted_solution_%s.pkl' % pipeline_id)
        if exists(fitted_pipeline_path):
            pipeline_path_dst = join(self.output_folder, 'temp', 'saved_fitted_pipelines', pipeline_id)
            create_outputfolders(pipeline_path_dst)
            shutil.copyfile(fitted_pipeline_path, join(pipeline_path_dst, 'fitted_solution_%s.pkl' % pipeline_id))
            return 'file://' + pipeline_path_dst

        return None

    def build_pipelines(self, session_id, dataset, task_keywords, metrics, timeout_search, timeout_run, hyperparameters,
                        template=None, targets=None, features=None, tune=None, report_rank=False):
        self.executor.submit(self._build_pipelines, session_id, dataset, task_keywords, metrics, timeout_search,
                             timeout_run, hyperparameters, template, targets, features, tune, report_rank)

    def build_fixed_pipeline(self, session_id, pipeline, dataset, targets=None, features=None):
        self.executor.submit(self._build_fixed_pipeline, session_id, pipeline, dataset, targets, features)

    # Runs in a worker thread from executor
    def _build_fixed_pipeline(self, session_id, pipeline_template, dataset, targets, features):

        session = self.sessions[session_id]
        with session.lock:
            # Force working=True so we get 'done_searching' even if no pipeline
            # gets created
            session.working = True

        db = self.DBSession()

        if dataset:
            dataset_uri = dataset
        else:
            dataset_uri = 'NA'

        pipeline_database = database.Pipeline(origin='Fixed pipeline template', dataset=dataset_uri)

        # TODO: Do it on d3mpipeline_generator.py
        def make_module(package, version, name):
            pipeline_module = database.PipelineModule(
                pipeline=pipeline_database,
                package=package, version=version, name=name)
            db.add(pipeline_module)
            return pipeline_module

        def make_data_module(name):
            return make_module('data', '0.0', name)

        def make_primitive_module(name):
            if name[0] == '.':
                name = 'd3m.primitives' + name
            return make_module('d3m', '2018.7.10', name)

        def connect(from_module, to_module,
                    from_output='produce', to_input='inputs'):
            db.add(database.PipelineConnection(pipeline=pipeline_database,
                                               from_module=from_module,
                                               to_module=to_module,
                                               from_output_name=from_output,
                                               to_input_name=to_input))

        def set_hyperparams(module, **hyperparams):
            db.add(database.PipelineParameter(
                pipeline=pipeline_database, module=module,
                name='hyperparams', value=pickle.dumps(hyperparams),
            ))

        try:
            # TODO: Use pipeline input for this
            if dataset:
                input_data = make_data_module('dataset')
                db.add(database.PipelineParameter(
                    pipeline=pipeline_database, module=input_data,
                    name='targets', value=pickle.dumps(targets),
                ))
                db.add(database.PipelineParameter(
                    pipeline=pipeline_database, module=input_data,
                    name='features', value=pickle.dumps(features),
                ))
            prev_step = None
            prev_steps = {}
            count_template_steps = 0
            for pipeline_step in pipeline_template['steps']:
                if pipeline_step['type'] == 'PRIMITIVE':
                    step = make_primitive_module(pipeline_step['primitive']['python_path'])
                    if 'outputs' in pipeline_step:
                        for output in pipeline_step['outputs']:
                            prev_steps['steps.%d.%s' % (count_template_steps,output['id'])] = step

                    count_template_steps += 1
                    if 'hyperparams' in pipeline_step:
                        hyperparams = {}
                        for hyper, desc in pipeline_step['hyperparams'].items():
                            hyperparams[hyper] = {'type':desc['type'] ,'data':desc['data']}
                        set_hyperparams(step, **hyperparams)
                else:
                    # TODO In the future we should be able to handle subpipelines
                    break
                if prev_step:
                    if 'arguments' in pipeline_step:
                        for argument, desc in pipeline_step['arguments'].items():
                            connect(prev_steps[desc['data']], step, from_output=desc['data'].split('.')[-1], to_input=argument)
                    connect(prev_step, step, from_output='index', to_input='index')
                else:
                    connect(input_data, step, from_output='dataset')
                prev_step = step
            db.add(pipeline_database)
            db.commit()
            pipeline_id = pipeline_database.id
            logger.info("Created fixed pipeline %s", pipeline_id)
            session.write_searched_pipeline(pipeline_id)

            session.notify('new_fixed_pipeline', pipeline_id=pipeline_id)
            with session.lock:
                session.pipelines.add(pipeline_id)
                # Force working=True so we get 'done_searching' even if no pipeline
                # gets created
                session.working = False
        finally:
            db.close()

    # Runs in a worker thread from executor
    def _build_pipelines(self, session_id, dataset_uri, task_keywords, metrics, timeout_search, timeout_run,
                         hyperparameters, pipeline_template, targets, features, tune, report_rank):
        """Generates pipelines for the session.
        """
        session = self.sessions[session_id]
        with session.lock:
            session.targets = targets
            session.features = features
            if session.metrics != metrics:
                if session.metrics:
                    old = 'from %s ' % ', '.join([m['metric'] for m in session.metrics])
                else:
                    old = ''
                session.metrics = metrics
                logger.info("Set metrics to %s %s(for session %s)", metrics, old, session_id)

            # Force working=True so we get 'done_searching' even if no pipeline
            # gets created
            session.working = True

        if timeout_search is None:
            timeout_search = MAX_RUNNING_TIME

        timeout_search = timeout_search * 60  # Minutes to seconds
        timeout_search_internal = timeout_search

        if PIPELINES_TO_TUNE > 0:
            timeout_search_internal = timeout_search * (1 - TIME_TO_TUNE)

        now = time.time()
        expected_search_end = now + timeout_search
        sample_dataset_uri = self._get_sample_uri(dataset_uri, session.problem)
        metadata = profile_data(dataset_uri, session.targets)
        time_already_used = time.time() - now  # Time already used for sampling and profiling
        timeout_search_internal = timeout_search_internal - time_already_used
        session.dataset_uri = dataset_uri
        session.sample_dataset_uri = sample_dataset_uri
        session.report_rank = report_rank
        session.timeout_run = timeout_run
        session.expected_search_end = expected_search_end

        if 'TEMPLATES' in SEARCH_STRATEGIES:
            self._build_pipeline_from_template(session, task_keywords, dataset_uri, sample_dataset_uri, metrics,
                                               hyperparameters, metadata, timeout_search_internal)
        if 'ALPHA_AUTOML' in SEARCH_STRATEGIES:
            self._build_pipelines_from_generator(session, task_keywords, dataset_uri, sample_dataset_uri, metrics,
                                                 hyperparameters, metadata, timeout_search_internal, pipeline_template)

        session.tune_when_ready(tune)

    def _build_pipelines_from_generator(self, session, task_keywords, dataset_uri, sample_dataset_uri, metrics,
                                        hyperparameters, metadata, timeout_search, pipeline_template):

        logger.info("Starting AlphaD3M process, timeout is %s", timeout_search)
        msg_queue = Receiver()
        proc = run_process(
            'alphad3m.pipeline_synthesis.setup_search.generate_pipelines',
            'generate',
            msg_queue,
            task_keywords=task_keywords,
            dataset=dataset_uri,
            metrics=metrics,
            problem=session.problem,
            targets=session.targets,
            features=session.features,
            hyperparameters=hyperparameters,
            metadata=metadata,
            pipeline_template=pipeline_template,
            time_bound=timeout_search,
            db_filename=self.db_filename,
        )

        start = time.time()
        stop = False

        # Now we wait for pipelines to be sent over the pipe
        while proc.poll() is None:
            if not stop:
                if session.stop_requested:
                    logger.error("Session stop requested, sending SIGTERM to generator process")
                    stop = True

                if time.time() > start + timeout_search:
                    logger.error("Reached search timeout (%d > %d seconds), sending SIGTERM to generator process",
                                 time.time() - start, timeout_search)
                    stop = True

                if stop:
                    proc.terminate()
                    try:
                        proc.wait(30)
                    except subprocess.TimeoutExpired:
                        proc.kill()
                        proc.wait()
            try:
                msg, *args = msg_queue.recv(3)
            except Empty:
                continue

            if msg == 'eval':
                pipeline_id, = args
                logger.info("Got pipeline %s from generator process", pipeline_id)
                score = self.run_pipeline(session, dataset_uri, sample_dataset_uri, task_keywords, pipeline_id)
                logger.info("Sending score to generator process")
                if not stop:
                    msg_queue.send(score)
            else:
                raise RuntimeError("Got unknown message from generator process: %r" % msg)

        logger.warning("Generator process exited with %r", proc.returncode)

    def _build_pipeline_from_template(self, session, task_keywords, dataset_uri, sample_dataset_uri, metrics,
                                      hyperparameters, metadata, timeout_search):

        pipeline_ids = generate_pipelines(task_keywords, dataset_uri, session.problem, session.targets, session.features,
                                          hyperparameters, metadata,  metrics, self.DBSession)
        for pipeline_id in pipeline_ids:
            if pipeline_id is not None:
                try:
                    # Add it to the session
                    session.add_scoring_pipeline(pipeline_id)
                    logger.info('Created pipeline %s', pipeline_id)
                    scoring_config = get_internal_scoring_config(task_keywords)

                    self._run_queue.put(ScoreJob(self, pipeline_id, dataset_uri, session.metrics, session.problem,
                                                 scoring_config, timeout_search, session.report_rank,
                                                 sample_dataset_uri))

                    session.notify('new_pipeline', pipeline_id=pipeline_id)
                except Exception:
                    logger.exception('Error building pipeline from template')

    def run_pipeline(self, session, dataset_uri, sample_dataset_uri, task_keywords, pipeline_id):

        """Score a single pipeline.

        This is used by the pipeline synthesis code.
        """
        scoring_config = get_internal_scoring_config(task_keywords)
        timeout_run = session.timeout_run

        if timeout_run is None:
            timeout_run = TIME_TO_SCORE
            if any(tk in [TaskKeyword.VIDEO, TaskKeyword.IMAGE, TaskKeyword.AUDIO] for tk in task_keywords):
                timeout_run = timeout_run * 2  # Increase the score timeout for video, image and audio datasets

        timeout_run = timeout_run * 60  # Minutes to seconds

        # Add the pipeline to the session, score it
        with session.with_observer_queue() as queue:
            session.add_scoring_pipeline(pipeline_id)
            logger.info("Created pipeline %s", pipeline_id)
            self._run_queue.put(ScoreJob(self, pipeline_id, dataset_uri, session.metrics, session.problem,
                                         scoring_config, timeout_run, session.report_rank, sample_dataset_uri))
            session.notify('new_pipeline', pipeline_id=pipeline_id)

            while True:
                event, kwargs = queue.get(True)
                if event == 'done_searching':
                    raise RuntimeError("Never got pipeline results")
                elif (event == 'scoring_error' and
                      kwargs['pipeline_id'] == pipeline_id):
                    return None
                elif (event == 'scoring_success' and
                      kwargs['pipeline_id'] == pipeline_id):
                    break

        db = self.DBSession()
        try:
            # Find most recent evaluation
            evaluation_id = (
                select([database.Evaluation.id])
                    .where(database.Evaluation.pipeline_id == pipeline_id)
                    .order_by(database.Evaluation.date.desc())
            ).as_scalar()
            # Get scores from that evaluation
            scores = db.query(
                select([func.avg(database.EvaluationScore.value),
                        database.EvaluationScore.metric])
                    .where(
                    database.EvaluationScore.evaluation_id ==
                    evaluation_id
                )
                    .group_by(database.EvaluationScore.metric)
            ).all()

            first_metric = session.metrics[0]['metric'].name
            for value, metric in scores:
                if metric == first_metric:
                    logger.info("Evaluation result: %s -> %r", metric, value)
                    return value
            logger.info("Didn't get the requested metric from evaluation")
            return None
        finally:
            db.close()

    def _get_sample_uri(self, dataset_uri, problem):
        logger.info('About to sample dataset %s', dataset_uri)
        task_keywords = problem['problem']['task_keywords']

        if any(tk in [TaskKeyword.OBJECT_DETECTION, TaskKeyword.FORECASTING, TaskKeyword.GRAPH] for tk in task_keywords):
            logger.info('Not doing sampling for task %s', '_'.join([x.name for x in task_keywords]))
            return None

        dataset = Dataset.load(dataset_uri)

        if is_collection(dataset_uri[7:]):
            logger.info('Not doing sampling for collections')
            return None

        dataset_sample_folder = 'file://%s/temp/dataset_sample/' % os.environ.get('D3MOUTPUTDIR')
        dataset_sample_uri = None

        if os.path.exists(dataset_sample_folder[6:]):
            shutil.rmtree(dataset_sample_folder[6:])

        dataset_sample = get_dataset_sample(dataset, problem, dataset_sample_folder)

        if isinstance(dataset_sample, str):  # Was the dataset sampled?
            dataset_sample_uri = dataset_sample

        return dataset_sample_uri

    # Runs in a background thread
    def _pipeline_running_thread(self):
        running_jobs = {}
        while True:
            # Poll jobs, remove finished ones
            remove = []
            for job in running_jobs.values():
                if job.check():
                    remove.append(id(job))
            for job_id in remove:
                del running_jobs[job_id]

            # Start new jobs if we are under the maximum
            if len(running_jobs) < MAX_RUNNING_PROCESSES:
                try:
                    job = self._run_queue.get(False)
                except Empty:
                    pass
                else:
                    job.start(db_filename=self.db_filename,
                              predictions_root=self.runtime_folder)
                    running_jobs[id(job)] = job

            time.sleep(3)


class Session(Observable):
    """A session, in the gRPC meaning.

    This corresponds to a search in which pipelines are created.
    """
    def __init__(self, ta2, problem, output_folder, DBSession):
        Observable.__init__(self)
        self.id = uuid4()
        self._ta2 = ta2
        self.problem = problem
        self.DBSession = DBSession

        # D3M folders
        str_id = str(self.id)
        self._searched_pipelines_dir = os.path.join(output_folder, str_id, 'pipelines_searched')
        self._scored_pipelines_dir = os.path.join(output_folder, str_id, 'pipelines_scored')
        self._ranked_pipelines_dir = os.path.join(output_folder, str_id, 'pipelines_ranked')
        self._run_pipelines_dir = os.path.join(output_folder, str_id, 'pipeline_runs')
        create_outputfolders(self._searched_pipelines_dir)
        create_outputfolders(self._scored_pipelines_dir)
        create_outputfolders(self._ranked_pipelines_dir)
        create_outputfolders(self._run_pipelines_dir)

        self.metrics = []
        self.report_rank = False

        self._observer = self._ta2.add_observer(self._ta2_event)

        self.start = datetime.datetime.utcnow()

        # Should tuning be triggered when we are done with current pipelines?
        self._tune_when_ready = None

        # All the pipelines that belong to this session
        self.pipelines = set()
        # The pipelines currently in the queue for scoring
        self.pipelines_scoring = set()
        # The pipelines in the queue for hyperparameter tuning
        self.pipelines_tuning = set()
        # Pipelines already tuned, and pipelines created through tuning
        self.tuned_pipelines = set()
        # Flag indicating we started scoring & tuning, and a
        # 'done_searching' signal should be sent once no pipeline is pending
        self.working = False
        # Flag allowing TA3 to stop the search early
        self.stop_requested = False

        # Read metrics from problem
        if self.problem is not None:
            self.metrics = problem['problem']['performance_metrics']

        self._targets = None
        self._features = None
        self.dataset_uri = None
        self.sample_dataset_uri = None
        self.timeout_run = None
        self.expected_search_end = None

    @property
    def problem_id(self):
        return self.problem['id']

    @property
    def targets(self):
        if self._targets is not None:
            return set(self._targets)
        else:
            # Read targets from problem
            targets = set()
            #assert len(self.problem['inputs']) == 1
            for target in self.problem['inputs'][0]['targets']:
                targets.add((target['resource_id'], target['column_name']))
            return targets

    @targets.setter
    def targets(self, value):
        if value is None:
            self._targets = None
        elif isinstance(value, (set, list)):
            if not value:
                raise ValueError("Can't set targets to empty set")
            self._targets = set(value)
        else:
            raise TypeError("targets should be a set or None")

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, value):
        if value is None:
            self._features = None
        elif isinstance(value, (set, list)):
            if not value:
                raise ValueError("Can't set features to empty set")
            self._features = set(value)
        else:
            raise TypeError("features should be a set or None")

    def _ta2_event(self, event, **kwargs):
        if event == 'scoring_start':
            if kwargs['pipeline_id'] in self.pipelines_scoring:
                logger.info("Scoring pipeline for %s (session %s has %d "
                            "pipelines left to score)",
                            kwargs['pipeline_id'], self.id,
                            len(self.pipelines_scoring))
                self.notify(event, **kwargs)
        elif event == 'scoring_success' or event == 'scoring_error':
            if kwargs['pipeline_id'] in self.pipelines_scoring:
                self.notify(event, **kwargs)
                self.pipeline_scoring_done(kwargs['pipeline_id'], event)

    def tune_when_ready(self, tune=None):
        if tune is None:
            tune = PIPELINES_TO_TUNE
        self._tune_when_ready = tune
        self.working = True
        self.check_status()

    def add_scoring_pipeline(self, pipeline_id):
        with self.lock:
            self.working = True
            self.pipelines.add(pipeline_id)
            self.pipelines_scoring.add(pipeline_id)
            self.write_searched_pipeline(pipeline_id)

    def pipeline_scoring_done(self, pipeline_id, event=None):
        with self.lock:
            self.pipelines_scoring.discard(pipeline_id)
            self.check_status()
            if event == 'scoring_success':
                self.write_scored_pipeline(pipeline_id)

    def pipeline_tuning_done(self, old_pipeline_id, new_pipeline_id=None):
        with self.lock:
            self.pipelines_tuning.discard(old_pipeline_id)
            self.tuned_pipelines.add(old_pipeline_id)
            if new_pipeline_id is not None:
                self.pipelines.add(new_pipeline_id)
                self.tuned_pipelines.add(new_pipeline_id)
                self.write_searched_pipeline(new_pipeline_id)
                self.write_scored_pipeline(new_pipeline_id)
            self.check_status()

    @property
    def progress(self):
        if self._tune_when_ready is not None:
            to_tune = self._tune_when_ready - len(self.tuned_pipelines) / 2
        else:
            to_tune = 0
        return ProgressStatus(
            current=len(self.pipelines) - len(self.pipelines_scoring),
            total=len(self.pipelines) + to_tune,
        )

    def get_top_pipelines(self, db, metric, limit=None):
        metric_name = metric.name
        pipeline = aliased(database.Pipeline)
        evaluation_score = (
            select([func.avg(database.EvaluationScore.value)])
            .where(database.EvaluationScore.evaluation_id ==
                   database.Evaluation.id)
            .where(database.EvaluationScore.metric == metric_name)
            .where(database.Evaluation.pipeline_id == pipeline.id)
            .as_scalar()
        )
        if metric.best_value() == 1:
            eval_score_order = evaluation_score.desc()
        else:
            eval_score_order = evaluation_score.asc()  # Error based metrics
        q = (
            db.query(pipeline, evaluation_score)
            .filter(pipeline.id.in_(self.pipelines))
            .filter(evaluation_score != None)
            # FIXME: Using a joined load here results in duplicated results
            .options(lazyload(pipeline.parameters))
            .order_by(eval_score_order)
        )
        if limit is not None:
            q = q.limit(limit)
        return q.all()

    def check_status(self):
        with self.lock:
            # Session is not to be finished automatically
            if self._tune_when_ready is None:
                return
            # We are already done
            if not self.working:
                return
            # If pipelines are still in the queue
            if self.pipelines_scoring or self.pipelines_tuning:
                return

            db = self.DBSession()
            try:
                # If we are out of pipelines to score, maybe submit pipelines for
                # tuning
                logger.info("Session %s: scoring done", self.id)

                tune = []
                if self._tune_when_ready and self.stop_requested:
                    logger.info("Session stop requested, skipping tuning")
                elif self._tune_when_ready:
                    top_pipelines = self.get_top_pipelines(
                        db, self.metrics[0]['metric'],
                        self._tune_when_ready)
                    for pipeline, _ in top_pipelines:
                        if pipeline.id not in self.tuned_pipelines:
                            tune.append(pipeline.id)

                    if len(tune) > 0:
                        # Found some pipelines to tune, do that
                        logger.warning("Found %d pipelines to tune", len(tune))
                        timeout_tuning = self.expected_search_end - time.time()
                        for pipeline_id in tune:
                            logger.info("    %s", pipeline_id)
                            self._ta2._run_queue.put(TuneHyperparamsJob(self._ta2, self, pipeline_id, self.problem, timeout_tuning))
                            self.pipelines_tuning.add(pipeline_id)
                        return
                    logger.info("Found no pipeline to tune")
                else:
                    logger.info("No tuning requested")

                # Session is done (but new pipelines might be added later)
                self.working = False
                self.notify('done_searching')

                logger.warning("Search done")
                if self.metrics:
                    metric = self.metrics[0]['metric']
                    top_pipelines = self.get_top_pipelines(db, metric)
                    logger.warning("Found %d pipelines", len(top_pipelines))

                    for i, (pipeline, score) in enumerate(top_pipelines):
                        created = pipeline.created_date - self.start
                        logger.info("    %d) %s %s=%s origin=%s time=%.2fs",
                                    i + 1, pipeline.id, metric, score,
                                    pipeline.origin, created.total_seconds())
            finally:
                db.close()

    def write_searched_pipeline(self, pipeline_id):
        if not self._searched_pipelines_dir:
            logger.info("Not writing log file")
            return

        db = self.DBSession()
        try:
            # Get pipeline
            pipeline = db.query(database.Pipeline).get(pipeline_id)

            logger.warning("Writing searched_pipeline JSON for pipeline %s "
                           "origin=%s",
                           pipeline_id, pipeline.origin)

            filename = os.path.join(self._searched_pipelines_dir,
                                    '%s.json' % pipeline_id)
            obj = to_d3m_json(pipeline)
            with open(filename, 'w') as fp:
                json.dump(obj, fp, indent=2)
        except Exception:
            logger.exception("Error writing searched_pipeline for %s",
                             pipeline_id)
        finally:
            db.close()

    def write_scored_pipeline(self, pipeline_id):
        if not self._scored_pipelines_dir:
            logger.info("Not writing log file")
            return

        db = self.DBSession()
        try:
            # Get pipeline
            pipeline = db.query(database.Pipeline).get(pipeline_id)

            logger.warning("Writing scored_pipeline JSON for pipeline %s "
                           "origin=%s",
                           pipeline_id, pipeline.origin)

            filename = os.path.join(self._scored_pipelines_dir,
                                    '%s.json' % pipeline_id)
            obj = to_d3m_json(pipeline)
            with open(filename, 'w') as fp:
                json.dump(obj, fp, indent=2)
        except Exception:
            logger.exception("Error writing scored_pipeline for %s",
                             pipeline_id)
        finally:
            db.close()

    def write_exported_pipeline(self, pipeline_id, rank=None):
        metric = self.metrics[0]['metric'].name

        db = self.DBSession()
        try:
            # Get pipeline
            pipeline = db.query(database.Pipeline).get(pipeline_id)

            if rank is None:
                # Find most recent evaluation
                evaluation_id = (
                    select([database.Evaluation.id])
                    .where(database.Evaluation.pipeline_id == pipeline_id)
                    .order_by(database.Evaluation.date.desc())
                ).as_scalar()
                # Get score from that evaluation
                score = db.query(
                    select([func.avg(database.EvaluationScore.value)])
                    .where(
                        database.EvaluationScore.evaluation_id ==
                        evaluation_id
                    )
                    .where(database.EvaluationScore.metric == metric)
                    .as_scalar()
                )
                if score is None:
                    rank = 1000.0
                    logger.error("Writing pipeline JSON for pipeline %s, but "
                                 "it is not scored for %s. Rank set to %s. "
                                 "origin=%s",
                                 pipeline_id, metric, rank, pipeline.origin)
                else:
                    logger.warning("Writing pipeline JSON for pipeline %s "
                                   "%s=%s origin=%s",
                                   pipeline_id, metric, score.value,
                                   pipeline.origin)
                    rank = 1.0 - self.metrics[0]['metric'].normalize(score.value)
            else:
                logger.warning("Writing pipeline JSON for pipeline %s with "
                               "provided rank %s. origin=%s",
                               pipeline_id, rank, pipeline.origin)

            obj = to_d3m_json(pipeline)

            with open(os.path.join(self._ranked_pipelines_dir, '%s.json' % pipeline_id), 'w') as fout:
                json.dump(obj, fout, indent=2)
            with open(os.path.join(self._ranked_pipelines_dir, '%s.rank' % pipeline_id), 'w') as fout:
                fout.write(str(rank))

        finally:
            db.close()

    def close(self):
        self._ta2.remove_observer(self._observer)
        self._observer = None
        self.stop_requested = True
        self.notify('finish_session')


class Job(object):
    def __init__(self):
        self.msg = None

    def start(self, **kwargs):
        raise NotImplementedError

    def check(self):
        if self.msg is not None:
            try:
                while True:
                    self.message(*self.msg.recv(0))
            except Empty:
                pass

        return self.poll()

    def poll(self):
        raise NotImplementedError

    def message(self, *args):
        raise NotImplementedError


class ScoreJob(Job):

    def __init__(self, ta2, pipeline_id, dataset_uri, metrics, problem, scoring_config, timeout_run, report_rank=False,
                 sample_dataset_uri=None):
        Job.__init__(self)
        self.ta2 = ta2
        self.pipeline_id = pipeline_id
        self.dataset_uri = dataset_uri
        self.sample_dataset_uri = sample_dataset_uri
        self.metrics = metrics
        self.problem = problem
        self.scoring_config = scoring_config
        self.timeout_run = timeout_run
        self.expected_end = time.time() + timeout_run
        self.report_rank = report_rank

    def start(self, db_filename, **kwargs):
        self.msg = Receiver()
        self.proc = run_process('alphad3m.pipeline_operations.pipeline_score.score', 'score', self.msg,
                                pipeline_id=self.pipeline_id,
                                dataset_uri=self.dataset_uri,
                                sample_dataset_uri=self.sample_dataset_uri,
                                metrics=self.metrics,
                                problem=self.problem,
                                scoring_config=self.scoring_config,
                                report_rank=self.report_rank,
                                db_filename=db_filename)

        self.ta2.notify('scoring_start',
                        pipeline_id=self.pipeline_id,
                        job_id=id(self))

    def poll(self):
        timeout_reached = False
        if time.time() > self.expected_end:
            timeout_reached = True
            logger.error('Reached timeout (%d seconds) to score a pipeline' % self.timeout_run)
            self.proc.terminate()
            try:
                self.proc.wait(30)
            except subprocess.TimeoutExpired:
                self.proc.kill()
                self.proc.wait()

        if self.proc.poll() is None:
            return False

        error_msg = read_streams(self.proc)
        if timeout_reached: error_msg = 'Reached timeout (%d seconds) to score a pipeline' % self.timeout_run
        log = logger.info if self.proc.returncode == 0 else logger.error
        log('Pipeline scoring process done, returned %d (pipeline: %s)', self.proc.returncode, self.pipeline_id)

        if self.proc.returncode == 0:
            self.ta2.notify('scoring_success',
                            pipeline_id=self.pipeline_id,
                            job_id=id(self))
        else:
            self.ta2.notify('scoring_error',
                            pipeline_id=self.pipeline_id,
                            job_id=id(self),
                            error_msg=error_msg)
        return True


class TrainJob(Job):
    def __init__(self, ta2, pipeline_id, dataset, problem, steps_to_expose):
        Job.__init__(self)
        self.ta2 = ta2
        self.pipeline_id = pipeline_id
        self.dataset = dataset
        self.problem = problem
        self.steps_to_expose = steps_to_expose

    def start(self, db_filename, **kwargs):
        logger.info('Training pipeline for %s', self.pipeline_id)
        self.msg = Receiver()
        self.proc = run_process('alphad3m.pipeline_operations.pipeline_train.train', 'train', self.msg,
                                pipeline_id=self.pipeline_id,
                                dataset=self.dataset,
                                problem=self.problem,
                                storage_dir=self.ta2.runtime_folder,
                                steps_to_expose=self.steps_to_expose,
                                db_filename=db_filename)
        self.ta2.notify('training_start',
                        pipeline_id=self.pipeline_id,
                        job_id=id(self))

    def poll(self):
        if self.proc.poll() is None:
            return False

        error_msg = read_streams(self.proc)
        log = logger.info if self.proc.returncode == 0 else logger.error
        log('Pipeline training process done, returned %d (pipeline: %s)', self.proc.returncode, self.pipeline_id)

        if self.proc.returncode == 0:
            steps_to_expose = []
            for step_id in self.steps_to_expose:
                if exists(join(self.ta2.runtime_folder, 'fit_%s_%s.csv' % (self.pipeline_id, step_id))):
                    steps_to_expose.append(step_id)
            self.ta2.notify('training_success',
                            pipeline_id=self.pipeline_id,
                            storage_dir=self.ta2.runtime_folder,
                            steps_to_expose=steps_to_expose,
                            job_id=id(self))
        else:
            self.ta2.notify('training_error',
                            pipeline_id=self.pipeline_id,
                            job_id=id(self),
                            error_msg=error_msg)
        return True


class TestJob(Job):
    def __init__(self, ta2, pipeline_id, dataset, steps_to_expose):
        Job.__init__(self)
        self.ta2 = ta2
        self.pipeline_id = pipeline_id
        self.dataset = dataset
        self.steps_to_expose = steps_to_expose

    def start(self, db_filename, **kwargs):
        logger.info("Testing pipeline for %s", self.pipeline_id)
        self.msg = Receiver()
        self.proc = run_process('alphad3m.pipeline_operations.pipeline_test.test', 'test', self.msg,
                                pipeline_id=self.pipeline_id,
                                dataset=self.dataset,
                                storage_dir=self.ta2.runtime_folder,
                                steps_to_expose=self.steps_to_expose,
                                db_filename=db_filename)
        self.ta2.notify('testing_start',
                        pipeline_id=self.pipeline_id,
                        job_id=id(self))

    def poll(self):
        if self.proc.poll() is None:
            return False

        error_msg = read_streams(self.proc)
        log = logger.info if self.proc.returncode == 0 else logger.error
        log('Pipeline testing process done, returned %d (pipeline: %s)', self.proc.returncode, self.pipeline_id)

        if self.proc.returncode == 0:
            steps_to_expose = []
            for step_id in self.steps_to_expose:
                if exists(join(self.ta2.runtime_folder, 'produce_%s_%s.csv' % (self.pipeline_id, step_id))):
                    steps_to_expose.append(step_id)
            self.ta2.notify('testing_success',
                            pipeline_id=self.pipeline_id,
                            storage_dir=self.ta2.runtime_folder,
                            steps_to_expose=steps_to_expose,
                            job_id=id(self))
        else:
            self.ta2.notify('testing_error',
                            pipeline_id=self.pipeline_id,
                            job_id=id(self),
                            error_msg=error_msg)
        return True


class TuneHyperparamsJob(Job):
    def __init__(self, ta2, session, pipeline_id, problem, timeout_tuning):
        Job.__init__(self)
        self.ta2 = ta2
        self.session = session
        self.pipeline_id = pipeline_id
        self.problem = problem
        self.timeout_tuning = timeout_tuning
        self.expected_end = time.time() + timeout_tuning

    def start(self, db_filename, **kwargs):
        logger.info("Running tuning for %s (session %s has %d pipelines left to tune)",
                    self.pipeline_id, self.session.id,
                    len(self.session.pipelines_tuning))

        self.msg = Receiver()

        self.proc = run_process('alphad3m.pipeline_operations.pipeline_tune.tune',
                                'tune', self.msg,
                                pipeline_id=self.pipeline_id,
                                storage_dir=self.ta2.internal_folder,
                                dataset_uri=self.session.dataset_uri,
                                sample_dataset_uri=self.session.sample_dataset_uri,
                                metrics=self.session.metrics,
                                problem=self.problem,
                                scoring_config=get_internal_scoring_config(self.problem['problem']['task_keywords']),
                                report_rank=self.session.report_rank,
                                timeout_tuning=self.timeout_tuning,
                                db_filename=db_filename)

        self.session.notify('tuning_start',
                            pipeline_id=self.pipeline_id,
                            job_id=id(self))

    def poll(self):
        if time.time() > self.expected_end:
            logger.error('Reached timeout (%d seconds) to tune a pipeline' % self.timeout_tuning)
            self.proc.terminate()
            try:
                self.proc.wait(30)
            except subprocess.TimeoutExpired:
                self.proc.kill()
                self.proc.wait()

        if self.proc.poll() is None:
            return False

        error_msg = read_streams(self.proc)
        log = logger.info if self.proc.returncode == 0 else logger.error
        log("Pipeline tuning process done, returned %d (pipeline: %s)",
            self.proc.returncode, self.pipeline_id)

        if self.proc.returncode == 0:
            logger.info("New pipeline: %s)", self.tuned_pipeline_id)
            self.session.notify('tuning_success',
                                old_pipeline_id=self.pipeline_id,
                                new_pipeline_id=self.tuned_pipeline_id,
                                job_id=id(self))
            self.session.notify('scoring_success',
                                pipeline_id=self.tuned_pipeline_id,
                                job_id=id(self))
            self.session.pipeline_tuning_done(self.pipeline_id,
                                              self.tuned_pipeline_id)
        else:
            self.session.notify('tuning_error',
                                pipeline_id=self.pipeline_id,
                                job_id=id(self),
                                error_msg=error_msg)
            self.session.pipeline_tuning_done(self.pipeline_id)
        return True

    def message(self, msg, arg):
        if msg == 'progress':
            # TODO: Report progress
            logger.info("Tuning pipeline %s: %.0f%%",
                        self.pipeline_id, arg * 100)
        elif msg == 'tuned_pipeline_id':
            self.tuned_pipeline_id = arg
        else:
            logger.error("Unexpected message from tuning process %s",
                         msg)


class ThreadPoolExecutor(futures.ThreadPoolExecutor):
    def submit(self, fn, *args, **kwargs):
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except Exception:
                logger.exception("Exception in worker thread")
                raise
        return futures.ThreadPoolExecutor.submit(self, wrapper,
                                                 *args, **kwargs)



import logging
import os
import sys
import shutil
import pickle
from os.path import join
from sqlalchemy.orm import joinedload
from alphad3m.pipeline_operations.pipeline_score import calculate_scores, save_scores
from alphad3m.schema import database
from alphad3m.hyperparameter_tuning.primitive_config import is_tunable
from alphad3m.hyperparameter_tuning.bayesian import HyperparameterTuning, get_new_hyperparameters
from alphad3m.utils import create_outputfolders

logger = logging.getLogger(__name__)


@database.with_db
def tune(pipeline_id, dataset_uri, sample_dataset_uri, storage_dir, metrics, problem, scoring_config, report_rank, timeout_tuning, msg_queue, db):
    # Load pipeline from database
    pipeline = (
        db.query(database.Pipeline)
        .filter(database.Pipeline.id == pipeline_id)
        .options(joinedload(database.Pipeline.modules),
                 joinedload(database.Pipeline.connections))
    ).one()

    timeout_tuning = timeout_tuning - 90  # Reduce 1.5 minutes to finish safely the tuning process
    logger.info('About to tune pipeline, id=%s, timeout=%d secs', pipeline_id, timeout_tuning)
    tunable_primitives = {}

    for primitive in pipeline.modules:
        if is_tunable(primitive.name):
            tunable_primitives[primitive.id] = primitive.name

    if len(tunable_primitives) == 0:
        logger.info('No primitives to be tuned for pipeline %s', pipeline_id)
        sys.exit(1)

    logger.info('Tuning primitives: %s', ', '.join(tunable_primitives.values()))

    def evaluate_tune(hyperparameter_configuration):
        new_hyperparams = []

        for primitive_id, primitive_name in tunable_primitives.items():
            hy = get_new_hyperparameters(primitive_name, hyperparameter_configuration)
            db_hyperparams = database.PipelineParameter(
                pipeline=pipeline,
                module_id=primitive_id,
                name='hyperparams',
                value=pickle.dumps(hy),
            )
            new_hyperparams.append(db_hyperparams)

        pipeline.parameters += new_hyperparams
        scores = calculate_scores(pipeline, dataset_uri, sample_dataset_uri, metrics, problem, scoring_config)
        first_metric = metrics[0]['metric'].name
        score_values = []

        for fold_scores in scores.values():
            for metric, score_value in fold_scores.items():
                if metric == first_metric:
                    score_values.append(score_value)

        avg_score = sum(score_values) / len(score_values)
        cost = 1.0 - metrics[0]['metric'].normalize(avg_score)
        logger.info('Tuning results:\n%s, cost=%s', scores, cost)

        return cost, scores

    # Run tuning, gets best configuration
    tuning = HyperparameterTuning(tunable_primitives.values())
    output_directory = join(storage_dir, 'tuning', str(pipeline_id))
    create_outputfolders(output_directory)
    best_configuration, best_scores = tuning.tune(evaluate_tune, timeout_tuning, output_directory)

    # Duplicate pipeline in database
    new_pipeline = database.duplicate_pipeline(db, pipeline, 'HyperparameterTuning from pipeline %s' % pipeline_id)

    for primitive in new_pipeline.modules:
        if is_tunable(primitive.name):
            best_hyperparameters = get_new_hyperparameters(primitive.name, best_configuration)
            query = db.query(database.PipelineParameter).filter(database.PipelineParameter.module_id == primitive.id)\
                .filter(database.PipelineParameter.pipeline_id == new_pipeline.id)\
                .filter(database.PipelineParameter.name == 'hyperparams')
            if query.first():
                original_parameters = pickle.loads(query.first().value)
                original_parameters.update(best_hyperparameters)
                query.update({database.PipelineParameter.value: pickle.dumps(original_parameters)})
            else:
                db.add(database.PipelineParameter(
                    pipeline=new_pipeline,
                    module_id=primitive.id,
                    name='hyperparams',
                    value=pickle.dumps(best_hyperparameters),
                ))

    save_scores(new_pipeline.id, best_scores, metrics, report_rank, db)
    shutil.rmtree(output_directory)
    logger.info('Tuning done, generated new pipeline %s', new_pipeline.id)
    msg_queue.send(('tuned_pipeline_id', new_pipeline.id))



class Game():
    """
    This class specifies the base Game class. To define your own game, subclass
    this class and implement the functions below. This works when the game is
    two-player, adversarial and turn-based.

    Use 1 for player1 and -1 for player2.

    See othello/OthelloGame.py for an example implementation.
    """
    def __init__(self):
        pass

    def getInitBoard(self):
        """
        Returns:
            startBoard: a representation of the board (ideally this is the form
                        that will be the input to your neural network)
        """
        pass

    def getBoardSize(self):
        """
        Returns:
            (x,y): a tuple of board dimensions
        """
        pass

    def getActionSize(self):
        """
        Returns:
            actionSize: number of all possible actions
        """
        pass

    def getNextState(self, board, player, action):
        """
        Input:
            board: current board
            player: current player (1 or -1)
            action: action taken by current player

        Returns:
            nextBoard: board after applying action
            nextPlayer: player who plays in the next turn (should be -player)
        """
        pass

    def getValidMoves(self, board, player):
        """
        Input:
            board: current board
            player: current player

        Returns:
            validMoves: a binary vector of length self.getActionSize(), 1 for
                        moves that are valid from the current board and player,
                        0 for invalid moves
        """
        pass

    def getGameEnded(self, board, player):
        """
        Input:
            board: current board
            player: current player (1 or -1)

        Returns:
            r: 0 if game has not ended. 1 if player won, -1 if player lost.
        """
        pass

    def getCanonicalForm(self, board, player):
        """
        Input:
            board: current board
            player: current player (1 or -1)

        Returns:
            canonicalBoard: returns canonical form of board. The canonical form
                            should be independent of player. For e.g. in chess,
                            the canonical form can be chosen to be from the pov
                            of white. When the player is white, we can return
                            board as is. When the player is black, we can invert
                            the colors and return the board.
        """
        pass

    def getSymmetries(self, board, pi):
        """
        Input:
            board: current board
            pi: policy vector of size self.getActionSize()

        Returns:
            symmForms: a list of [(board,pi)] where each tuple is a symmetrical
                       form of the board and the corresponding pi vector. This
                       is used when training the neural network from examples.
        """
        pass

    def stringRepresentation(self, board):
        """
        Input:

            board: current board

        Returns:
            boardString: a quick conversion of board to a string format.
                         Required by MCTS for hashing.
        """
        pass



import argparse
import os
import shutil
import time
import random
import numpy as np
import sys
import logging

#from alphad3m.pipeline_search.utils import Bar, AverageMeter
from alphad3m.pipeline_search.NeuralNet import NeuralNet

import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd import Variable

from alphad3m.pipeline_search.pipeline.PipelineNNet import PipelineNNet as onnet

logger = logging.getLogger(__name__)

args = dict({
    'lr': 0.001,
    'dropout': 0.3,
    'epochs': 2,
    'batch_size': 64,
    'cuda': False,
    'num_channels': 512,
})

class NNetWrapper(NeuralNet):
    def __init__(self, game):
        self.nnet = onnet(game, args)
        self.action_size = game.getActionSize()
        self.board_size = game.getBoardSize()
        if args.get('cuda'):
            self.nnet.cuda()

    def train(self, examples):
        """
        examples: list of examples, each example is of form (board, pi, v)
        """
        optimizer = optim.Adam(self.nnet.parameters())

        for epoch in range(args.get('epochs')):
            logger.info('EPOCH ::: %s', str(epoch+1))
            self.nnet.train()
            #data_time = AverageMeter()
            #batch_time = AverageMeter()
            #pi_losses = AverageMeter()
            #v_losses = AverageMeter()
            #end = time.time()

            batch_size = args.get('batch_size')
            #bar = Bar('Training Net', max=int(len(examples)/batch_size))
            batch_idx = 0

            while batch_idx < int(len(examples)/batch_size):
                sample_ids = np.random.randint(len(examples), size=batch_size)
                boards, pis, vs = list(zip(*[examples[i] for i in sample_ids]))
                boards = torch.FloatTensor(np.array(boards).astype(np.float64))
                target_pis = torch.FloatTensor(np.array(pis))
                target_vs = torch.FloatTensor(np.array(vs).astype(np.float64))

                #  predict
                if args.get('cuda'):
                    boards, target_pis, target_vs = Variable(boards.contiguous().cuda(),requires_grad=True), Variable(target_pis.contiguous().cuda(), requires_grad=True), Variable(target_vs.contiguous().cuda(),requires_grad=True)
                else:
                    boards, target_pis, target_vs = Variable(boards), Variable(target_pis),  Variable(target_vs)


                # measure data loading time
                #data_time.update(time.time() - end)

                # compute output
                #print(boards)
                out_pi, out_v = self.nnet(boards)
                l_pi = self.loss_pi(target_pis, out_pi)
                l_v = self.loss_v(target_vs, out_v)
                total_loss = l_pi + l_v

                # record loss
                #pi_losses.update(l_pi.data, boards.size(0))
                #v_losses.update(l_v.data, boards.size(0))

                # compute gradient and do SGD step
                optimizer.zero_grad()
                total_loss.backward()
                optimizer.step()

                # measure elapsed time
                #batch_time.update(time.time() - end)
                end = time.time()
                batch_idx += 1

                # plot progress
                #bar.suffix  = '({batch}/{size}) Data: {data:.3f}s | Batch: {bt:.3f}s | Total: {total:} | ETA: {eta:} | Loss_pi: {lpi:.4f} | Loss_v: {lv:.3f}'.format(
                #            batch=batch_idx,
                #            size=int(len(examples)/batch_size),
                #            data=data_time.avg,
                #            bt=batch_time.avg,
                #            total=bar.elapsed_td,
                #            eta=bar.eta_td,
                #            lpi=pi_losses.avg,
                #            lv=v_losses.avg,
                #            )
                #bar.next()
            #bar.finish()


    def predict(self, board):
        """
        board: np array with board
        """
        # timing
        start = time.time()
        #print('BOARD\n', board)

        # preparing input
        board = torch.from_numpy(np.array(board[0:self.board_size], dtype='f')).cuda().float() if args.get('cuda') else torch.from_numpy(np.array(board[0:self.board_size], dtype='f'))
        #board = torch.FloatTensor(board[0:self.board_x])
        if args.get('cuda'): board = board.contiguous().cuda()
        board = Variable(board, volatile=True)
        board = board.view(1, self.board_size)

        self.nnet.eval()
        pi, v = self.nnet(board)

        #print('PROBABILITY ', torch.exp(pi).data.cpu().numpy()[0])
        #print('VALUE ',  v.data.cpu().numpy()[0])

        #logger.info('PREDICTION TIME TAKEN : {0:03f}'.format(time.time()-start))
        return torch.exp(pi).data.cpu().numpy()[0], v.data.cpu().numpy()[0][0]

    def loss_pi(self, targets, outputs):
        return -torch.sum(targets*outputs)/targets.size()[0]

    def loss_v(self, targets, outputs):
        return torch.sum((targets-outputs.view(-1))**2)/targets.size()[0]

    def save_checkpoint(self, folder='checkpoint', filename='checkpoint.pth.tar'):
        filepath = os.path.join(folder, filename)
        if not os.path.exists(folder):
            logger.warning("Checkpoint Directory does not exist! Making directory {}".format(folder))
            os.mkdir(folder)
        else:
            logger.info("Checkpoint Directory exists! ")
        torch.save({
            'state_dict' : self.nnet.state_dict(),
        }, filepath)

    def load_checkpoint(self, folder='checkpoint', filename='checkpoint.pth.tar'):
        # https://github.com/pytorch/examples/blob/master/imagenet/main.py#L98
        filepath = os.path.join(folder, filename)
        if not os.path.exists(filepath):
            raise("No model in path {}".format(folder))

        checkpoint = torch.load(filepath)
        self.nnet.load_state_dict(checkpoint['state_dict'])
